from typing import List, Optional, Union

import pandas as pd
from sgqlc.operation import Operation

from ML_management.executor import BaseExecutor
from ML_management.model.model_type_to_methods_map import ModelMethodName
from ML_management.sdk import schema
from ML_management.sdk.model import get_latest_model_version
from ML_management.sdk.parameters import ArbitraryModels, ModelMethodParams, ModelVersionChoice, ModelWithRole
from ML_management.sdk.schema import Executor, ExecutorVersion
from ML_management.sdk.sdk import _generate_fake_schema, _print_params_by_schema, _to_datetime, send_graphql_request


def list_executor() -> pd.DataFrame:
    """
    List available executors.

    Returns
    -------
    pd.DataFrame
        Pandas dataframe with list of available executors.
    """
    op = Operation(schema.Query)
    op.list_executor.name()
    op.list_executor.description()
    op.list_executor.creation_timestamp()
    op.list_executor.last_updated_timestamp()
    json_data = send_graphql_request(op)
    df = pd.DataFrame.from_dict(json_data["listExecutor"])
    if not df.empty:
        df = _to_datetime(df, ["creationTimestamp", "lastUpdatedTimestamp"])
    return df


def set_executor_tag(name: str, key: str, value: str) -> Executor:
    """
    Set executor tag.

    Parameters
    ----------
    name: str
        Name of the model.
    key: str
        Key tag.
    value: str
        Value tag.

    Returns
    -------
    Executor
        Executor instance with meta information.
    """
    op = Operation(schema.Mutation)
    set_tag = op.set_executor_tag(name=name, key=key, value=value)
    set_tag.name()
    set_tag.tags()
    executor = send_graphql_request(op=op)
    return Executor(executor)


def delete_executor_tag(name: str, key: str) -> Executor:
    """
    Delete executor tag.

    Parameters
    ----------
    name: str
        Name of the model.
    key: str
        Key tag.

    Returns
    -------
    Executor
        Executor instance with meta information.
    """
    op = Operation(schema.Mutation)
    delete_tag = op.delete_executor_tag(name=name, key=key)
    delete_tag.name()
    delete_tag.tags()
    executor = send_graphql_request(op=op)
    return Executor(executor)


def set_executor_description(name: str, description: str) -> Executor:
    """
    Set executor description.

    Parameters
    ----------
    name: str
        Name of the model.
    description: str
        Description model.

    Returns
    -------
    Executor
        Executor instance with meta information.
    """
    op = Operation(schema.Mutation)
    set_description = op.set_executor_description(name=name, description=description)
    set_description.name()
    set_description.description()
    executor = send_graphql_request(op=op)
    return Executor(executor)


def delete_executor(executor_name: str) -> bool:
    """
    Delete executor and all of it's versions.

    Parameters
    ----------
    executor_name: str
        Name of the executor to delete.

    Returns
    -------
    bool
        Operation success status.
    """
    op = Operation(schema.Mutation)
    op.delete_executor(name=executor_name)
    return send_graphql_request(op)["deleteExecutor"]


def list_executor_version(name: str) -> pd.DataFrame:
    """
    List available versions of the executor with such name.

    Parameters
    ----------
    name: str
        Name of the executor.

    Returns
    -------
    pd.DataFrame
        Pandas dataframe with a list of available executor versions.
    """
    op = Operation(schema.Query)
    base_query = op.executor_from_name(name=name).list_executor_version
    base_query.version()
    base_query.creation_timestamp()
    base_query.name()
    json_data = send_graphql_request(op)

    df = pd.DataFrame.from_dict(json_data["executorFromName"]["listExecutorVersion"])
    df = _to_datetime(df, ["creationTimestamp"])

    return df.sort_values(by=["version"], ignore_index=True)


def print_model_schema_for_executor(
    executor_name: str, models: List[dict], executor_version: Optional[int] = None
) -> None:
    """
    Print model schema for particular executor.

    Parameters
    ----------
    executor_name: str
        Name of the executor.
    models: List[dict]
        Necessary information about the model.

        Example1::

            [
                {"name": "simple_model"} # str
            ]

        Example2::

            [
                {
                    "name": "model_name1", # str
                    "role": "role1", # Optional[str]
                    "version": version # Optional[int]
                },
                {
                    "name": "model_name2", # str
                    "role": "role2", # Optional[str]
                    "version": version # Optional[int]
                }
            ]
    executor_version: Optional[int] = None
        Version of the executor. Default: None, "latest" version is used.
    """
    models_methods_schemas = _get_model_schema_for_executor(
        executor_name=executor_name,
        executor_version=executor_version,
        models=models,
    )
    for model_methods_schemas in models_methods_schemas:
        role = model_methods_schemas["role"]
        for model in models:
            if ("role" in model and model["role"] == role) or "role" not in model:
                model_name = model["name"]

        print(f"Model name: {model_name}, model role: {role}")
        for model_methods_schema in model_methods_schemas["listMethodSchemas"]:
            _print_params_by_schema(
                model_methods_schema["jsonSchema"], ModelMethodName(model_methods_schema["schemaName"]).name
            )
        print()


def cancel_build_job_for_executor_version(executor_name: str, executor_version: int) -> bool:
    """
    Cancel running or planned build job of executor's image.

    Parameters
    ----------
    executor_name: str
        The name of the executor.
    executor_version: int
        The version of the executor.

    Returns
    -------
    bool
        Operation success status.
    """
    op = Operation(schema.Mutation)
    op.cancel_build_job_for_executor_version(name=executor_name, version=executor_version)
    return send_graphql_request(op)["cancelBuildJobForExecutorVersion"]


def delete_executor_version(executor_name: str, executor_version: int) -> bool:
    """
    Delete version of a executor.

    Parameters
    ----------
    executor_name: str
        The name of the executor.
    executor_version: int
        The version of the executor.

    Returns
    -------
    bool
        Operation success status.
    """
    op = Operation(schema.Mutation)
    executor_version_choice = schema.ObjectVersionInput(name=executor_name, version=executor_version)
    op.delete_executor_version_from_name_version(executor_version=executor_version_choice)
    return send_graphql_request(op)["deleteExecutorVersionFromNameVersion"]


def _get_model_schema_for_executor(executor_name: str, executor_version: Optional[int], models: list) -> dict:
    role_models = []

    for model in models:
        model_name = model["name"]
        model_version = model.get("version", None)
        if model_version is None:
            model_version = get_latest_model_version(model_name).version
        model_role = model.get("role", BaseExecutor.DEFAULT_ROLE)

        current_model = schema.ObjectVersionInput(name=model_name, version=model_version)
        current_role = schema.RoleObjectVersionInput(role=model_role, object_version=current_model)

        role_models.append(current_role)

    op = Operation(schema.Query)

    _executor_version = schema.ObjectVersionOptionalInput(name=executor_name, version=executor_version)
    base_query = (
        op.executor_version_from_name_version(executor_version=_executor_version)
        .job_json_schema_for_models(models=role_models)
        .list_role_model_method_schemas
    )

    base_query.role()
    base_query.list_method_schemas.schema_name()
    base_query.list_method_schemas.json_schema()

    json_data = send_graphql_request(op)
    return json_data["executorVersionFromNameVersion"]["jobJsonSchemaForModels"]["listRoleModelMethodSchemas"]


def generate_model_params_for_executor(
    executor_name: str, models: List[dict], executor_version: Optional[int] = None
) -> Union[List[ModelMethodParams], ArbitraryModels]:
    """
    Return example of model's methods parameters for executor.

    Parameters
    ----------
    executor_name: str
        Name of the executor.
    models: List[dict]
        Necessary information about the model.

        Example1::

            [
                {"name": "simple_model"}
            ];

        Example2::

            [
                {
                    "name": "model_name1", # str
                    "role": "role1", # Optional[str]
                    "version": version # Optional[int]
                },
                {
                    "name": "model_name2", # str
                    "role": "role2", # Optional[str]
                    "version": version # Optional[int]
                }
            ]

    executor_version: Optional[int] = None
        Version of the executor. Default: None, "latest" version is used.

    Returns
    -------
    List[ModelMethodParams]:
        Example of model's methods parameters in case simple executor.
        Example output::

            [
                ModelMethodParams(
                    method=ModelMethodName.evaluate_function,
                    params={key1: value1, key2: value2}
                )
            ]

    ArbitraryModels:
        Example of models parameter for add_ml_job function.
        Example::

            ArbitraryModels(
                models=[
                    ModelWithRole(
                        role="role_one",
                        model_version_choice=ModelVersionChoice(
                            name="model_1_name",
                            version=1
                        )
                        params=[
                            ModelMethodParams(
                                method=ModelMethodName.evaluate_function,
                                params={key1: value1, key2: value2}
                            ),
                            ModelMethodParams(
                                method=ModelMethodName.finetune_function,
                                params={key1: value1, key2: value2}
                            )
                        ]
                    ),
                    ModelWithRole(
                        role="role_two",
                        model_version_choice=ModelVersionChoice(
                            name="model_2_name",
                            version=1
                        )
                        params=[
                            ModelMethodParams(
                                method=ModelMethodName.evaluate_function,
                                params={key1: value1, key2: value2}
                            ),
                            ModelMethodParams(
                                method=ModelMethodName.finetune_function,
                                params={key1: value1, key2: value2}
                            )
                        ]
                    ),
                ]
            )

    """
    models_methods_schemas = _get_model_schema_for_executor(
        executor_name=executor_name,
        executor_version=executor_version,
        models=models,
    )

    sdk_models = []

    for model_methods_schemas in models_methods_schemas:
        list_model_params = []
        for model_methods_schema in model_methods_schemas["listMethodSchemas"]:
            list_model_params.append(
                ModelMethodParams(
                    method=ModelMethodName(model_methods_schema["schemaName"]),
                    params=_generate_fake_schema(model_methods_schema["jsonSchema"]),
                )
            )
        if len(models_methods_schemas) == 1:
            return list_model_params
        role = model_methods_schemas["role"]
        for model in models:
            if model["role"] == role:
                model_name = model["name"]
                model_version = model.get("version", None)
                break
        current_model = ModelWithRole(
            role=role,
            model_version_choice=ModelVersionChoice(name=model_name, version=model_version),
            params=list_model_params,
        )

        sdk_models.append(current_model)

    return ArbitraryModels(models=sdk_models)


def print_executor_schema(name: str, version: Optional[int] = None) -> None:
    """
    Print executor schema.

    Parameters
    ----------
    name: str
        Name of the executor.
    version: Optional[int] = None
        Version of the executor. Default: None, "latest" version is used.
    """
    op = Operation(schema.Query)
    executor_version = schema.ObjectVersionOptionalInput(name=name, version=version)
    base_query = op.executor_version_from_name_version(executor_version=executor_version)
    base_query.executor_method_schema()
    json_data = send_graphql_request(op)

    json_data = json_data["executorVersionFromNameVersion"]["executorMethodSchema"]
    _print_params_by_schema(json_schema=json_data, schema_type="Executor")


def print_executor_roles(name: str, version: Optional[int] = None) -> None:
    """
    Print the roles required by the executor.

    Parameters
    ----------
    name: str
        Name of the executor.
    version: Optional[int] = None
        Version of the executor. Default: None, "latest" version is used.
    """
    op = Operation(schema.Query)
    executor_version = schema.ObjectVersionOptionalInput(name=name, version=version)
    base_query = op.executor_version_from_name_version(executor_version=executor_version)
    base_query.desired_model_methods()
    base_query.desired_dataset_loader_methods()
    json_data = send_graphql_request(op)
    print("Desired model methods:", json_data["executorVersionFromNameVersion"]["desiredModelMethods"])
    print("Desired dataset loader methods:", json_data["executorVersionFromNameVersion"]["desiredDatasetLoaderMethods"])


def get_required_classes_by_executor(name: str, version: Optional[int] = None) -> List[str]:
    """
    Return the names of classes for the model to be inherited from, by the name of the executor.

    Parameters
    ----------
    name: str
        Name of the executor.
    executor_version: Optional[int] = None
        Version of the executor. Default: None, "latest" version is used.

    Returns
    -------
    List[str]:
        List of model class names to be inherited from.
    """
    op = Operation(schema.Query)
    executor_version = schema.ObjectVersionOptionalInput(name=name, version=version)
    base_query = op.executor_version_from_name_version(executor_version=executor_version)
    base_query.desired_model_patterns()
    json_data = send_graphql_request(op)
    return json_data["executorVersionFromNameVersion"]["desiredModelPatterns"]


def set_executor_version_description(name: str, version: int, description: str) -> ExecutorVersion:
    """
    Set executor version description.

    Parameters
    ----------
    name: str
        Name of the executor.
    version: int
        Version of the executor.
    description: str
        Description executor version.

    Returns
    -------
    ExecutorVersion
        Executor version instance with meta information.
    """
    op = Operation(schema.Mutation)
    choice = schema.ObjectVersionInput(name=name, version=version)
    set_description = op.set_executor_version_description(executor_version=choice, description=description)
    set_description.name()
    set_description.version()
    set_description.description()

    executor = send_graphql_request(op=op)
    return ExecutorVersion(executor)


def set_executor_version_tag(name: str, version: int, key: str, value: str) -> ExecutorVersion:
    """
    Set executor version tag.

    Parameters
    ----------
    name: str
        Name of the executor.
    version: int
        Version of the executor.
    key: str
        Key tag.
    value: str
        Value tag.

    Returns
    -------
    ExecutorVersion
        Executor version instance with meta information.
    """
    op = Operation(schema.Mutation)
    choice = schema.ObjectVersionInput(name=name, version=version)
    set_tag = op.set_executor_version_tag(executor_version=choice, key=key, value=value)
    set_tag.name()
    set_tag.version()
    set_tag.tags()
    executor = send_graphql_request(op=op)
    return ExecutorVersion(executor)


def delete_executor_version_tag(name: str, version: int, key: str) -> ExecutorVersion:
    """
    Delete executor version tag.

    Parameters
    ----------
    name: str
        Name of the executor.
    version: int
        Version of the executor.
    key: str
        Key tag.

    Returns
    -------
    ExecutorVersion
        Executor version instance with meta information.
    """
    op = Operation(schema.Mutation)
    choice = schema.ObjectVersionInput(name=name, version=version)
    delete_tag = op.delete_executor_version_tag(executor_version=choice, key=key)
    delete_tag.name()
    delete_tag.version()
    delete_tag.tags()
    executor = send_graphql_request(op=op)
    return ExecutorVersion(executor)


def get_executor_version(name: str, version: Optional[int] = None) -> ExecutorVersion:
    """
    Meta information about the executor version by the executor name and version.

    Parameters
    ----------
    name: str
        Name of the executor.
    version: Optional[int] = None
        Version of the executor. Default: None, "latest" version is used.

    Returns
    -------
    ExecutorVersion
        ExecutorVersion instance with meta information.
    """
    op = Operation(schema.Query)
    executor_version_choice = schema.ObjectVersionOptionalInput(name=name, version=version)
    base_query = op.executor_version_from_name_version(executor_version=executor_version_choice)
    base_query.name()
    base_query.version()
    base_query.run.run_id()
    base_query.tags()
    base_query.description()
    base_query.creation_timestamp()
    data = send_graphql_request(op, json_response=False).executor_version_from_name_version
    return data


def get_executor_version_conda_env(name: str, version: int) -> dict:
    """
    Condas configuration for the executor version by the executor name and version.

    Parameters
    ----------
    name: str
        Name of the executor.
    version: Optional[int] = None
        Version of the executor. Default: None, "latest" version is used.

    Returns
    -------
    Dict
        Dict with conda configuration.
    """
    op = Operation(schema.Query)
    _model_version = schema.ObjectVersionOptionalInput(name=name, version=version)
    base_query = op.executor_version_from_name_version(executor_version=_model_version)
    base_query.get_conda_env()
    model_version = send_graphql_request(op, json_response=False)
    return model_version.executor_version_from_name_version.get_conda_env


def get_executor_version_requirements(name: str, version: int) -> list:
    """
    Requirements for the executor version by the executor name and version.

    Parameters
    ----------
    name: str
        Name of the executor.
    version: Optional[int] = None
        Version of the executor. Default: None, "latest" version is used.

    Returns
    -------
    List
        List of requirements.
    """
    op = Operation(schema.Query)
    _model_version = schema.ObjectVersionOptionalInput(name=name, version=version)
    base_query = op.executor_version_from_name_version(executor_version=_model_version)
    base_query.list_requirements()
    model_version = send_graphql_request(op, json_response=False)
    return model_version.executor_version_from_name_version.list_requirements



"""This module create and send request to MLManagement server."""
import atexit
import inspect
from contextlib import _GeneratorContextManager
from typing import Any, Dict, Optional

import cloudpickle

from ML_management.mlmanagement import variables
from ML_management.mlmanagement.base_exceptions import MLMServerError
from ML_management.mlmanagement.session import AuthSession
from ML_management.mlmanagement.variables import (
    active_run_stack,
    get_server_ml_api,
)
from ML_management.mlmanagement.visibility_options import VisibilityOptions
from mlflow import ActiveRun
from mlflow.entities import RunStatus
from mlflow.exceptions import MlflowException, RestException
from mlflow.models import Model
from mlflow.tracking import _RUN_ID_ENV_VAR
from mlflow.utils import env


def create_kwargs(frame):
    """Get name and kwargs of function by its frame."""
    function_name = inspect.getframeinfo(frame)[2]  # get name of function
    _, _, _, kwargs = inspect.getargvalues(frame)  # get kwargs of function
    kwargs.pop("self", None)
    kwargs.pop("parts", None)
    kwargs.pop("python_path", None)

    return (
        function_name,
        kwargs,
    )  # return name of mlflow function and kwargs for that function


def request(
    function_name,
    kwargs,
    url=None,
    stream=False,
) -> _GeneratorContextManager:
    """Create mlflow_request and send it to server."""
    mlflow_request = {
        "function_name": function_name,
        "kwargs": kwargs,
        "active_experiment": variables.active_experiment,
        "active_run_ids": [run.info.run_id for run in active_run_stack],
    }

    url = url if url is not None else get_server_ml_api()

    return AuthSession().post(url=url, stream=stream, json=mlflow_request)


def send_request_to_server(function_name, kwargs):
    """
    Send request to server.

    Takes frame of mlflow func and extra_attr
    extra_attr is needed if original mlflow function is in the mlflow.<extra_attr> package
    for example function log_model is in mlflow.pyfunc module (mlflow.pyfunc.log_model())
    """
    with request(function_name, kwargs) as response:
        response_content = response.content

        try:
            decoded_result = cloudpickle.loads(response_content)
        except Exception:
            raise MLMServerError(f"Server error: {response_content.decode()}, status: {response.status_code}") from None

        # raise error if mlflow is supposed to raise error
        if isinstance(decoded_result, MlflowException):
            is_rest = decoded_result.json_kwargs.pop("isRest", False)
            if is_rest:
                created_json = {
                    "error_code": decoded_result.error_code,
                    "message": decoded_result.message,
                }
                decoded_result = RestException(created_json)
            raise decoded_result
        elif isinstance(decoded_result, Exception):
            raise decoded_result
        return decoded_result


def _check_if_call_from_predict_function():
    """
    Check if call to server was from predict function of model.

    Calls from predict function are prohibited and will do and return nothing.
    """
    from ML_management.model.model_type_to_methods_map import ModelMethodName
    from ML_management.model.patterns.model_pattern import Model

    predict_func_name = ModelMethodName.predict_function.name

    for frame in inspect.stack():
        if frame.function == predict_func_name and Model in frame[0].f_locals.get("self").__class__.__mro__:
            return True
    return False


def request_for_function(frame):
    """
    Send request to server or call mlflow function straightforward.

    Input parameters:
    :param frame: frame of equivalent mlflow function
    :param extra_attrs: list of extra modules for mlflow library, for example "tracking" (mlflow.tracking)
    :param class_name: the name of the class whose function is being called
    """
    if _check_if_call_from_predict_function():
        return None

    function_name, kwargs = create_kwargs(frame)

    return send_request_to_server(function_name, kwargs)


def set_experiment(
    experiment_name: str,
    visibility: VisibilityOptions = VisibilityOptions.PRIVATE,
) -> dict:
    """
    Set the given experiment as the active experiment.

    The experiment must either be specified by name via
    experiment_name.
    Set global variable active_experiment_name to that experiment_name.
    """
    variables.active_experiment = request_for_function(inspect.currentframe())
    return variables.active_experiment


def start_run(
    experiment_id: Optional[str] = None,
    run_name: Optional[str] = None,
    nested: bool = False,
    tags: Optional[Dict[str, Any]] = None,
) -> ActiveRun:
    """
    Start a new MLflow run, setting it as the active run under which metrics and parameters will be logged.

    The return value can be used as a context manager within a with block; otherwise, you must call end_run() to
    terminate the current run.
    If you pass a run_id or the MLFLOW_RUN_ID environment variable is set, start_run attempts to resume a run with
    the specified run ID and other parameters are ignored. run_id takes precedence over MLFLOW_RUN_ID.
    If resuming an existing run, the run status is set to RunStatus.RUNNING.
    Add that created run to active_run_stack.
    """
    if len(active_run_stack) > 0 and not nested:
        raise Exception(
            (
                "Run with UUID {} is already active. To start a new run, first end the "
                "current run with mlmanagement.end_run(). To start a nested "
                "run, call start_run with nested=True"
            ).format(active_run_stack[0].info.run_id)
        )
    _active_run = request_for_function(inspect.currentframe())
    active_run_stack.append(_active_run)
    return _active_run


def active_run() -> Optional[ActiveRun]:
    """Get the currently active Run, or None if no such run exists."""
    return active_run_stack[-1] if len(active_run_stack) > 0 else None


finished_run_status = RunStatus.to_string(RunStatus.FINISHED)


def end_run(status: str = finished_run_status) -> None:
    """End an active MLflow run (if there is one)."""
    if len(active_run_stack) > 0:
        # Clear out the global existing run environment variable as well.
        env.unset_variable(_RUN_ID_ENV_VAR)
        run = active_run_stack[-1]
        set_terminated(run.info.run_id, status)
        active_run_stack.pop()


atexit.register(end_run)


def set_terminated(run_id: str, status: Optional[str] = None, end_time: Optional[int] = None) -> None:
    """Set a run’s status to terminated."""
    return request_for_function(
        inspect.currentframe(),
    )


default_model = Model()


def _exit(self, exc_type, exc_val, exc_tb):
    """Redefine __exit__ function of class ActiveRun."""
    status = RunStatus.FINISHED if exc_type is None else RunStatus.FAILED
    end_run(RunStatus.to_string(status))
    return exc_type is None


# Rewrite __exit__ method to enable using Python ``with`` syntax of ActiveRun class.
ActiveRun.__exit__ = _exit


def start_run_if_not_exist():
    """If run doesn't exist call start_run() function."""
    if len(active_run_stack) == 0:
        start_run()



"""DatasetLoader template for dataset with base splits."""
from abc import abstractmethod

from ML_management.dataset_loader.dataset_loader_pattern import DatasetLoaderPattern


class BaseSplitsDatasetLoader(DatasetLoaderPattern):
    """Implementation of the DatasetLoader that performs basic splitting of data into training/validation/test parts."""

    def __init__(self):
        """Init dataset loader class."""
        super().__init__()

    @abstractmethod
    def get_train_data(self, **kwargs):
        """Define get_train_data function."""
        raise NotImplementedError

    @abstractmethod
    def get_validation_data(self, **kwargs):
        """Define get_validation_data function."""
        raise NotImplementedError

    @abstractmethod
    def get_test_data(self, **kwargs):
        """Define get_test_data function."""
        raise NotImplementedError



"""Base classes for custom exceptions used in project."""


class MLMBaseError(Exception):
    """Base exception for all custom exceptions."""

    pass


class MLMServerError(MLMBaseError):
    """Base exception for all server mlmanager exceptions."""

    pass


class RegistryError(MLMBaseError):
    """Base exception for all registry exceptions."""

    pass


class MLMClientError(MLMBaseError):
    """Base exception for all client-side specific exceptions."""

    pass


class PylintError(MLMBaseError):
    """Base exception for all errors within linter check."""

    def __init__(self, message):
        super().__init__("Pylint found errors in code of uploaded modules:\n" + message)



"""Custom exception definition for server mlflow manager graph."""

from typing import Any, Optional, Tuple

from ML_management.mlmanagement.base_exceptions import MLMServerError


class InvalidEnumTypeError(MLMServerError):
    """Define InvalidEnumTypeError exception."""

    def __init__(self, passed_enum_values, enum_type_name):
        self.passed_enum_values = passed_enum_values
        self.enum_type_name = enum_type_name
        self.message = (
            f'Passed enum values "{", ".join(passed_enum_values)}" is invalid, must be value of Enum {enum_type_name}.'
        )

        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (InvalidEnumTypeError, (self.passed_enum_values, self.enum_type_name))


class AuthError(MLMServerError):
    """Define AuthError exception."""

    def __init__(
        self, access_type: str, object_type: str, object_name: str, object_version: Optional[int] = None
    ) -> None:
        self.access_type = access_type
        self.object_type = object_type
        self.object_name = object_name
        self.object_version = object_version
        self.message = (
            f"Object {'version ' if object_version is not None else ''}access permission denied. "
            f"Access type: {access_type}, object type: {object_type}, object name: {object_name}"
        )
        if object_version is not None:
            self.message += f", object version: {object_version}"
        super().__init__(self.message)

    def __reduce__(self) -> Tuple[Any, ...]:
        res = (AuthError, (self.access_type, self.object_type, self.object_name))
        if self.object_version is not None:
            res = (res[0], res[1] + (self.object_version,))
        return res

    @property
    def params(self):
        return {
            "access_type": self.access_type,
            "object_type": self.object_type,
            "object_name": self.object_name,
            "object_version": self.object_version,
        }


class UnregisteredModelNameError(MLMServerError):
    """Define UnregisteredModelNameError exception."""

    def __init__(self, passed_name, kwarg):
        self.passed_name = passed_name
        self.kwarg = kwarg
        self.message = (
            'Passed "{kwarg}" argument value "{passed_name}" is invalid, as this model is not registered in mlflow.'
        ).format(
            kwarg=kwarg,
            passed_name=passed_name,
        )

        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (UnregisteredModelNameError, (self.passed_name, self.kwarg))


class ModelNamePatternViolationError(MLMServerError):
    """Define ModelNamePatternViolationError exception."""

    def __init__(self, passed_name, pattern):
        self.passed_name = passed_name
        self.pattern = pattern
        self.message = (
            "Model can not be registered with name {passed_name}, "
            "it must be not empty string and consist of alphanumeric "
            "characters, '_' and must start and end with an alphanumeric character."
            "Validation regexp: {pattern}"
        ).format(passed_name=passed_name, pattern=pattern)

        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (ModelNamePatternViolationError, (self.passed_name, self.pattern))


class InvalidModelVersionError(MLMServerError):
    """Define InvalidModelVersionError exception."""

    def __init__(self, passed_name, passed_version, name_kwarg, version_kwarg):
        self.passed_version = passed_version
        self.passed_name = passed_name
        self.name_kwarg = name_kwarg
        self.version_kwarg = version_kwarg
        self.message = 'Passed "{version_kwarg}" argument value "{passed_version}" is invalid, \
            as model "{passed_name}" (passed as "{name_kwarg}" argument value) has no such version.'.format(
            passed_name=passed_name,
            passed_version=passed_version,
            name_kwarg=name_kwarg,
            version_kwarg=version_kwarg,
        )

        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (
            InvalidModelVersionError,
            (
                self.passed_name,
                self.passed_version,
                self.name_kwarg,
                self.version_kwarg,
            ),
        )


class KwargNotPassedWithUploadTypeError(MLMServerError):
    """Define KwargNotPassedWithUploadTypeError exception."""

    def __init__(self, kwarg, passed_upload_model_type):
        self.passed_upload_model_type = passed_upload_model_type
        self.kwarg = kwarg
        self.message = (
            'Argument "{kwarg}" cannot be ommitted with upload_model_type UploadModelType.{passed_upload_model_type}.'
        ).format(
            kwarg=kwarg,
            passed_upload_model_type=passed_upload_model_type,
        )

        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (KwargNotPassedWithUploadTypeError, (self.kwarg, self.passed_upload_model_type))


class ModelTypeIsNotFoundError(MLMServerError):
    """Define ModelTypeIsNotFoundError exception."""

    def __init__(self):
        self.message = "Model type is not found. You must inherit the desired template."

        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (ModelTypeIsNotFoundError, ())


class ExistingModelWithOtherTypeError(MLMServerError):
    """Define ExistingModelWithOtherTypeError exception."""

    def __init__(self, name):
        self.name = name
        self.message = (
            f'The other model type with name "{name}" exists.'
            f"You cannot upload a model with the same name as an existing model of a different type."
        )

        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (ExistingModelWithOtherTypeError, (self.name,))


class InvalidExperimentNameError(MLMServerError):
    """Define InvalidExperimentNameError exception."""

    def __init__(self, model_type, exp_name):
        self.model_type = model_type
        self.exp_name = exp_name
        self.message = f"You can't specify '{exp_name}' experiment name for model type '{model_type}' upload."

        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (InvalidExperimentNameError, (self.model_type, self.exp_name))


class InvalidUploadModelModeError(MLMServerError):
    """Define InvalidUploadModelModeError exception."""

    def __init__(self, model_type, upload_mode):
        self.model_type = model_type
        self.upload_mode = upload_mode
        self.message = f"You can't specify upload mode '{upload_mode}' for '{model_type}' model."

        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (InvalidUploadModelModeError, (self.model_type, self.upload_mode))


class InvalidVisibilityOptionError(MLMServerError):
    """Define InvalidVisibilityOptionError exception."""

    def __init__(self, model_type):
        self.model_type = model_type
        self.message = f"You must specify visibility option for '{model_type}'."

        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (InvalidUploadModelModeError, (self.model_type,))



"""Custom exception definition (necessary for RegistryManager)."""

from ML_management.mlmanagement.base_exceptions import RegistryError


class VersionNotFoundError(RegistryError):
    """Define Version Not Found Exception."""

    def __init__(self, model_name: str, version: int, model_type: str = "model"):
        self.model_name = model_name
        self.model_type = model_type
        self.version = version
        self.message = f'There is no version {self.version} for {self.model_type} "{self.model_name}"'
        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (VersionNotFoundError, (self.model_name, self.version, self.model_type))


class MetricNotLoggedError(RegistryError):
    """Define Metric Not Logged exception."""

    def __init__(self, model_name: str, metric: str):
        self.model_name = model_name
        self.metric = metric
        self.message = f'Metric "{self.metric}" is not logged for model "{self.model_name}"'
        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (MetricNotLoggedError, (self.model_name, self.metric))


class ModelNotRegisteredError(RegistryError):
    """Define Model Not Registered exception."""

    def __init__(self, model_name: str, model_type: str = "model"):
        self.model_name = model_name
        self.model_type = model_type
        self.message = f'{model_type} "{model_name}" is not registered'
        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (ModelNotRegisteredError, (self.model_name, self.model_type))


class NoMetricProvidedError(RegistryError):
    """Define NoMetricProvidedError exception."""

    def __init__(self, criteria: str):
        self.criteria = criteria
        self.message = f'Choice criteria "{self.criteria}" is passed, but no metric name is provided'
        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (NoMetricProvidedError, (self.criteria))


class UnsupportedCriteriaError(RegistryError):
    """Define Unsupported Criteria exception."""

    def __init__(self, criteria: str, supported_criteria: list):
        self.criteria = criteria
        self.supported_criteria = supported_criteria
        self.message = f'Choice criteria "{self.criteria}" is unsupported, must be one of: {self.supported_criteria}'
        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (UnsupportedCriteriaError, (self.criteria, self.supported_criteria))


class ExperimentNotFoundNameError(RegistryError):
    """Define Experiment not found exception."""

    def __init__(self, experiment_name: str):
        self.experiment_name = experiment_name
        self.message = f'There is no experiment with name "{experiment_name}"'
        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (ExperimentNotFoundNameError, (self.experiment_name,))


class ExperimentNotFoundIDError(RegistryError):
    """Define Experiment not found exception."""

    def __init__(self, experiment_id: str):
        self.experiment_name = experiment_id
        self.message = f'There is no experiment with id "{experiment_id}"'
        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (ExperimentNotFoundIDError, (self.experiment_id,))


class JobNotFoundNameError(RegistryError):
    """Define job not found exception."""

    def __init__(self, job_name: str):
        self.job_name = job_name
        self.message = f'There is no job with name "{job_name}"'
        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (JobNotFoundNameError, (self.job_name,))


class JobHasNoRunIdError(RegistryError):
    """Define job has no run id exception."""

    def __init__(self, job_name: str):
        self.job_name = job_name
        self.message = f'Job with name "{job_name}" has no run. Maybe it is not started yet.'
        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (JobHasNoRunIdError, (self.job_name,))



from ML_management.sdk import dataset_loader, executor, experiment, job, model, parameters

list_model = model.list_model
list_dataset_loader = dataset_loader.list_dataset_loader
list_executor = executor.list_executor
add_ml_job = job.add_ml_job
job_by_name = job.job_by_name
job_metric_by_name = job.job_metric_by_name
list_model_version = model.list_model_version
list_dataset_loader_version = dataset_loader.list_dataset_loader_version
list_executor_version = executor.list_executor_version
model_version_metainfo = model.model_version_metainfo
rebuild_model_version_image = model.rebuild_model_version_image
print_model_schema_for_executor = executor.print_model_schema_for_executor
generate_model_params_for_executor = executor.generate_model_params_for_executor
print_dataset_loader_schema = dataset_loader.print_dataset_loader_schema
print_executor_schema = executor.print_executor_schema
print_executor_roles = executor.print_executor_roles
get_logs = job.get_logs
get_build_logs = job.get_build_logs
get_venv_build_logs = job.get_venv_build_logs
serve_model = model.serve_model
stop_model_serving = model.stop_model_serving
check_inference_model_readiness = model.check_inference_model_readiness
get_required_classes_by_executor = executor.get_required_classes_by_executor
available_metrics = job.available_metrics
metric_history = job.metric_history
list_params_job = job.list_params_job
delete_model_version = model.delete_model_version
delete_model = model.delete_model
delete_executor_version = executor.delete_executor_version
delete_executor = executor.delete_executor
delete_dataset_loader = dataset_loader.delete_dataset_loader
delete_dataset_loader_version = dataset_loader.delete_dataset_loader_version
cancel_job = job.cancel_job
cancel_build_job_for_model_version = model.cancel_build_job_for_model_version
cancel_build_job_for_executor_version = executor.cancel_build_job_for_executor_version
cancel_venv_build_job_for_model_version = model.cancel_venv_build_job_for_model_version
set_experiment_tag = experiment.set_experiment_tag
set_experiment_description = experiment.set_experiment_description
delete_experiment_tag = experiment.delete_experiment_tag
get_experiment_by_id = experiment.get_experiment_by_name
get_experiment_by_name = experiment.get_experiment_by_name
ModelMethodParams = parameters.ModelMethodParams
ModelVersionChoice = parameters.ModelVersionChoice
SingleModel = parameters.SingleModel
ModelWithRole = parameters.ModelWithRole
ArbitraryModels = parameters.ArbitraryModels
DatasetLoaderMethodParams = parameters.DatasetLoaderMethodParams
SingleDatasetLoader = parameters.SingleDatasetLoader
DatasetLoaderWithRole = parameters.DatasetLoaderWithRole
ArbitraryDatasetLoaders = parameters.ArbitraryDatasetLoaders
get_dataset_loader_version_requirements = dataset_loader.get_dataset_loader_version_requirements
get_executor_version_requirements = executor.get_executor_version_requirements
get_model_version_requirements = model.get_model_version_requirements
get_dataset_loader_version_conda_env = dataset_loader.get_dataset_loader_version_conda_env
get_executor_version_conda_env = executor.get_executor_version_conda_env
get_executor_version = executor.get_executor_version
get_model_version_conda_env = model.get_model_version_conda_env
get_dataset_loader_version = dataset_loader.get_dataset_loader_version
delete_dataset_loader_version_tag = dataset_loader.delete_dataset_loader_version_tag
delete_executor_version_tag = executor.delete_executor_version_tag
delete_model_version_tag = model.delete_model_version_tag
set_dataset_loader_version_tag = dataset_loader.set_dataset_loader_version_tag
set_executor_version_tag = executor.set_executor_version_tag
set_model_version_tag = model.set_model_version_tag
set_dataset_loader_version_description = dataset_loader.set_dataset_loader_version_description
set_executor_version_description = executor.set_executor_version_description
set_model_version_description = model.set_model_version_description
set_model_description = model.set_model_description
set_executor_description = executor.set_executor_description
set_dataset_loader_description = dataset_loader.set_dataset_loader_description
delete_model_tag = model.delete_model_tag
delete_executor_tag = executor.delete_executor_tag
delete_dataset_loader_tag = dataset_loader.delete_dataset_loader_tag
set_dataset_loader_tag = dataset_loader.set_dataset_loader_tag
set_executor_tag = executor.set_executor_tag
set_model_tag = model.set_model_tag
get_latest_model_version = model.get_latest_model_version
get_initial_model_version = model.get_initial_model_version



"""Custom _load_pyfunc function for loading model from source code."""
import importlib
import sys
from pathlib import Path

CONFIG_KEY_ARTIFACTS = "artifacts"


class PythonModelContext:
    """
    A collection of artifacts that a :class:`~PythonModel` can use when performing inference.

    :class:`~PythonModelContext` objects are created *implicitly* by the
    :func:`save_model() <mlflow.pyfunc.save_model>` and
    :func:`log_model() <mlflow.pyfunc.log_model>` persistence methods, using the contents specified
    by the ``artifacts`` parameter of these methods.
    """

    def __init__(self, artifacts):
        """:param artifacts: A dictionary of ``<name, artifact_path>``."""
        self._artifacts = artifacts

    @property
    def artifacts(self):
        """A dictionary containing ``<name, artifact_path>``."""
        return self._artifacts


def _load_pyfunc(model_data):
    from ML_management.mlmanagement.utils import INIT_FUNCTION_NAME  # circular import

    # structure of the model path /"job" + /model role name + /"data" +
    # + /name of the model folder (name of the folder from which model was originally logged)
    model_path = model_data["model_path"]
    if not model_data["kwargs_for_init"]:
        model_data["kwargs_for_init"] = {}
    parts = Path(model_path).parts
    if str(Path(*parts[:2])) not in sys.path:
        sys.path.append(str(Path(*parts[:2])))
    python_model = getattr(importlib.import_module(".".join(parts[2:])), INIT_FUNCTION_NAME)(
        **model_data["kwargs_for_init"]
    )
    artifacts = Path(model_path) / CONFIG_KEY_ARTIFACTS
    if not artifacts.exists():
        artifacts.mkdir()

    context = PythonModelContext(artifacts=str(artifacts))
    python_model.load_context(context=context)
    return _PythonModelPyfuncWrapper(python_model=python_model, context=context, signature=None)


class _PythonModelPyfuncWrapper:
    """Wrapper class."""

    def __init__(self, python_model, context, signature):
        """Init wrapper class.

        :param python_model: An instance of a subclass of :class:`~PythonModel`.
        :param context: A :class:`~PythonModelContext` instance containing artifacts that
                        ``python_model`` may use when performing inference.
        :param signature: :class:`~ModelSignature` instance describing model input and output.
        """
        self.python_model = python_model
        self.context = context
        self.signature = signature

    def predict(self, model_input):
        return self.python_model.predict_function(self.context, model_input)



from typing import Optional

import pandas as pd
from sgqlc.operation import Operation

from ML_management.dataset_loader.dataset_loader_pattern_to_methods_map import DatasetLoaderMethodName
from ML_management.sdk import schema
from ML_management.sdk.schema import DatasetLoader, DatasetLoaderVersion
from ML_management.sdk.sdk import _print_params_by_schema, _to_datetime, send_graphql_request


def list_dataset_loader() -> pd.DataFrame:
    """
    List available dataset_loaders.

    Returns
    -------
    pd.DataFrame
        Pandas dataframe with list of available dataset_loaders.
    """
    op = Operation(schema.Query)
    op.list_dataset_loader.name()
    op.list_dataset_loader.description()
    op.list_dataset_loader.creation_timestamp()
    op.list_dataset_loader.last_updated_timestamp()
    json_data = send_graphql_request(op)
    df = pd.DataFrame.from_dict(json_data["listDatasetLoader"])
    if not df.empty:
        df = _to_datetime(df, ["creationTimestamp", "lastUpdatedTimestamp"])
    return df


def set_dataset_loader_tag(name: str, key: str, value: str) -> DatasetLoader:
    """
    Set dataset loader tag.

    Parameters
    ----------
    name: str
        Name of the model.
    key: str
        Key tag.
    value: str
        Value tag.

    Returns
    -------
    DatasetLoader
        DatasetLoader instance with meta information.
    """
    op = Operation(schema.Mutation)
    set_tag = op.set_dataset_loader_tag(name=name, key=key, value=value)
    set_tag.name()
    set_tag.tags()
    dataset_loader = send_graphql_request(op=op)
    return DatasetLoader(dataset_loader)


def delete_dataset_loader_tag(name: str, key: str) -> DatasetLoader:
    """
    Delete dataset loader tag.

    Parameters
    ----------
    name: str
        Name of the model.
    key: str
        Key tag.

    Returns
    -------
    DatasetLoader
        DatasetLoader instance with meta information.
    """
    op = Operation(schema.Mutation)
    delete_tag = op.delete_dataset_loader_tag(name=name, key=key)
    delete_tag.name()
    delete_tag.tags()
    dataset_loader = send_graphql_request(op=op)
    return DatasetLoader(dataset_loader)


def set_dataset_loader_description(name: str, description: str) -> DatasetLoader:
    """
    Set dataset loader description.

    Parameters
    ----------
    name: str
        Name of the model.
    description: str
        Description model.

    Returns
    -------
    DatasetLoader
        DatasetLoader instance with meta information.
    """
    op = Operation(schema.Mutation)
    set_description = op.set_dataset_loader_description(name=name, description=description)
    set_description.name()
    set_description.description()
    dataset_loader = send_graphql_request(op=op)
    return DatasetLoader(dataset_loader)


def list_dataset_loader_version(name: str) -> pd.DataFrame:
    """
    List available versions of the dataset_loader with such name.

    Parameters
    ----------
    name: str
        Name of the DatasetLoader.

    Returns
    -------
    pd.DataFrame
        Pandas dataframe with a list of available dataset_loader versions.
    """
    op = Operation(schema.Query)
    base_query = op.dataset_loader_from_name(name=name).list_dataset_loader_version
    base_query.version()
    base_query.creation_timestamp()
    base_query.name()
    json_data = send_graphql_request(op)

    df = pd.DataFrame.from_dict(json_data["datasetLoaderFromName"]["listDatasetLoaderVersion"])
    df = _to_datetime(df, ["creationTimestamp"])

    return df.sort_values(by=["version"], ignore_index=True)


def delete_dataset_loader(dataset_loader_name: str) -> bool:
    """
    Delete dataset loader and all of it's versions.

    Parameters
    ----------
    dataset_loader_name: str
        Name of the dataset loader to delete.

    Returns
    -------
    bool
        Operation success status.
    """
    op = Operation(schema.Mutation)
    op.delete_dataset_loader(name=dataset_loader_name)
    return send_graphql_request(op)["deleteDatasetLoader"]


def delete_dataset_loader_version(dataset_loader_name: str, dataset_loader_version: int) -> bool:
    """
    Delete version of a dataset loader.

    Parameters
    ----------
    dataset_loader_name: str
        The name of the dataset loader.
    dataset_loader_version: int
        The version of the dataset loader.

    Returns
    -------
    bool
        Operation success status.
    """
    op = Operation(schema.Mutation)
    dataset_loader_version_choice = schema.ObjectVersionInput(name=dataset_loader_name, version=dataset_loader_version)
    op.delete_dataset_loader_version_from_name_version(dataset_loader_version=dataset_loader_version_choice)
    return send_graphql_request(op)["deleteDatasetLoaderVersionFromNameVersion"]


def print_dataset_loader_schema(name: str, version: Optional[int] = None) -> None:
    """
    Print DatasetLoader schema.

    Parameters
    ----------
    name: str
        Name of the DatasetLoader.
    version: Optional[int] = None
        Version of the DatasetLoader. Default: None, "latest" version is used.
    """
    op = Operation(schema.Query)
    _datasetloader_version = schema.ObjectVersionOptionalInput(name=name, version=version)
    base_query = op.dataset_loader_version_from_name_version(dataset_loader_version=_datasetloader_version)
    base_query.dataset_loader_method_schemas()
    json_data = send_graphql_request(op)
    json_data = json_data["datasetLoaderVersionFromNameVersion"]["datasetLoaderMethodSchemas"]
    print(f"DatasetLoader {name} version {version} json-schema:")
    for method_name, schema_ in json_data.items():
        _print_params_by_schema(json_schema=schema_, schema_type=DatasetLoaderMethodName(method_name).name)


def set_dataset_loader_version_description(name: str, version: int, description: str) -> DatasetLoaderVersion:
    """
    Set dataset loader version description.

    Parameters
    ----------
    name: str
        Name of the dataset loader.
    version: int
        Version of the dataset loader.
    description: str
        Description dataset loader version.

    Returns
    -------
    DatasetLoaderVersion
        Dataset loader version instance with meta information.
    """
    op = Operation(schema.Mutation)
    choice = schema.ObjectVersionInput(name=name, version=version)
    set_description = op.set_dataset_loader_version_description(dataset_loader_version=choice, description=description)
    set_description.name()
    set_description.version()
    set_description.description()

    dataset_loader = send_graphql_request(op=op)
    return DatasetLoaderVersion(dataset_loader)


def set_dataset_loader_version_tag(name: str, version: int, key: str, value: str) -> DatasetLoaderVersion:
    """
    Set dataset loader version tag.

    Parameters
    ----------
    name: str
        Name of the dataset loader.
    version: int
        Version of the dataset loader.
    key: str
        Key tag.
    value: str
        Value tag.

    Returns
    -------
    DatasetLoaderVersion
        Dataset loader version instance with meta information.
    """
    op = Operation(schema.Mutation)
    choice = schema.ObjectVersionInput(name=name, version=version)
    set_tag = op.set_dataset_loader_version_tag(dataset_loader_version=choice, key=key, value=value)
    set_tag.name()
    set_tag.version()
    set_tag.tags()
    dataset_loader = send_graphql_request(op=op)
    return DatasetLoaderVersion(dataset_loader)


def delete_dataset_loader_version_tag(name: str, version: int, key: str) -> DatasetLoaderVersion:
    """
    Delete dataset loader version tag.

    Parameters
    ----------
    name: str
        Name of the dataset loader.
    version: int
        Version of the dataset loader.
    key: str
        Key tag.

    Returns
    -------
    DatasetLoaderVersion
        Dataset loader version instance with meta information.
    """
    op = Operation(schema.Mutation)
    choice = schema.ObjectVersionInput(name=name, version=version)
    delete_tag = op.delete_dataset_loader_version_tag(dataset_loader_version=choice, key=key)
    delete_tag.name()
    delete_tag.version()
    delete_tag.tags()
    dataset_loader = send_graphql_request(op=op)
    return DatasetLoaderVersion(dataset_loader)


def get_dataset_loader_version(name: str, version: Optional[int] = None) -> DatasetLoaderVersion:
    """
    Meta information about the dataset loader version by the dataset loader name and version.

    Parameters
    ----------
    name: str
        Name of the model.
    version: Optional[int] = None
        Version of the dataset loader. Default: None, "latest" version is used.

    Returns
    -------
    DatasetLoaderVersion
        DatasetLoaderVersion instance with meta information.
    """
    op = Operation(schema.Query)
    dataset_loader_version_choice = schema.ObjectVersionOptionalInput(name=name, version=version)
    base_query = op.dataset_loader_version_from_name_version(dataset_loader_version=dataset_loader_version_choice)
    base_query.name()
    base_query.version()
    base_query.run.run_id()
    base_query.tags()
    base_query.description()
    base_query.creation_timestamp()
    data = send_graphql_request(op, json_response=False)
    return data.dataset_loader_version_from_name_version


def get_dataset_loader_version_conda_env(name: str, version: int) -> dict:
    """
    Condas configuration for the dataset loader version by the dataset loader name and version.

    Parameters
    ----------
    name: str
        Name of the dataset loader.
    version: Optional[int] = None
        Version of the dataset loader. Default: None, "latest" version is used.

    Returns
    -------
    Dict
        Dict with conda configuration.
    """
    op = Operation(schema.Query)
    _model_version = schema.ObjectVersionOptionalInput(name=name, version=version)
    base_query = op.dataset_loader_version_from_name_version(dataset_loader_version=_model_version)
    base_query.get_conda_env()
    model_version = send_graphql_request(op, json_response=False)
    return model_version.dataset_loader_version_from_name_version.get_conda_env


def get_dataset_loader_version_requirements(name: str, version: int) -> list:
    """
    Requirements for the dataset loader version by the dataset loader name and version.

    Parameters
    ----------
    name: str
        Name of the dataset loader.
    version: Optional[int] = None
        Version of the dataset loader. Default: None, "latest" version is used.

    Returns
    -------
    List
        List of requirements.
    """
    op = Operation(schema.Query)
    _model_version = schema.ObjectVersionOptionalInput(name=name, version=version)
    base_query = op.dataset_loader_version_from_name_version(dataset_loader_version=_model_version)
    base_query.list_requirements()
    model_version = send_graphql_request(op, json_response=False)
    return model_version.dataset_loader_version_from_name_version.list_requirements



"""S3 Manager for operations with s3 data."""
import asyncio
import os
import posixpath
import random
import threading
from contextlib import asynccontextmanager
from http import HTTPStatus
from typing import Any, AsyncGenerator, Coroutine, List, Optional, Union

import aioboto3
import boto3
from aiofiles import os as aos
from botocore.awsrequest import AWSRequest, AWSResponse
from botocore.exceptions import ClientError
from tqdm.autonotebook import tqdm

from ML_management.mlmanagement import variables
from ML_management.mlmanagement.load_api import _untar_folder
from ML_management.mlmanagement.log_api import _tar_folder
from ML_management.mlmanagement.session import AuthSession
from ML_management.mlmanagement.visibility_options import VisibilityOptions
from ML_management.s3.utils import get_upload_paths

MAX_OBJECTS_PER_PAGE = 50
MAX_TASKS_NUMBER = 50
MAX_DELETE_CONCURRENT_REQUESTS = 100


class S3BucketNotFoundError(Exception):
    """Define Bucket Not Found Exception."""

    def __init__(self, bucket: str):
        self.bucket = bucket
        self.message = f'Bucket "{bucket}" does not exist'
        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (S3BucketNotFoundError, (self.bucket,))


class S3ObjectNotFoundError(Exception):
    """Define Version Not Found Exception."""

    def __init__(self, path: str, bucket: str):
        self.path = path
        self.bucket = bucket
        self.message = f'Object "{path}" is not found in "{bucket}" bucket'
        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (S3ObjectNotFoundError, (self.path, self.bucket))


class FileIsNotTarError(Exception):
    """Define exception when tar is expected but not given."""

    def __init__(self, filename: str):
        self.filename = filename
        self.message = f"Expected tar file, but file {self.filename} does not have '.tar' extension"
        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (FileIsNotTarError, (self.filename,))


class AmbiguousFileChoiceError(Exception):
    """Define exception when there is not exactly one file in bucket to choice."""

    def __init__(self, number_of_files: int):
        self.number_of_files = number_of_files
        if not number_of_files:
            appendix = " bucket is empty"
        else:
            appendix = "there is more than one file in bucket. Specify one file name"
        self.message = f"Expected one tar file in bucket, but {appendix}."
        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (AmbiguousFileChoiceError, (self.number_of_files,))


class S3Manager:
    """Manager for operations with s3 objects."""

    def __init__(self) -> None:
        """Init creds."""
        self.default_url = variables.get_s3_gateway_url()
        self.default_access_key_id, self.default_secret_access_key = variables.get_s3_credentials()
        self.session = AuthSession()

    def list_buckets(self) -> List[str]:
        """Get list names of available buckets."""
        with self.session.get(posixpath.join(self.default_url, "list-buckets")) as response:
            buckets_info: list[dict[str, str]] = response.json()["buckets"]
            return [bucket["name"] for bucket in buckets_info]

    def create_bucket(self, name: str, visibility: VisibilityOptions) -> None:
        """Create new bucket with specified name and visibility."""
        with self.session.post(
            posixpath.join(self.default_url, "create-bucket"), json={"name": name, "visibility": visibility.value}
        ) as response:
            if response.status_code != HTTPStatus.CREATED:
                raise RuntimeError(f"Failed to create bucket: {response.text}")

    def update_bucket_visibility(self, bucket: str, new_visibility: VisibilityOptions) -> None:
        """Update bucket visibility."""
        if bucket not in self.list_buckets():
            raise S3BucketNotFoundError(bucket)
        with self.session.post(
            posixpath.join(self.default_url, "update-bucket-auth", bucket),
            json={"visibility": new_visibility.value},
        ) as response:
            response.raise_for_status()

    def delete_bucket(self, bucket: str) -> None:
        """Delete bucket, this function is for synchronous code only.

        If you want delete bucket inside async application, use `delete_bucket_async`
        """
        self.delete_by_prefix(bucket=bucket, prefix="")
        s3_client = self._get_sync_boto_client()
        s3_client.delete_bucket(Bucket=bucket)

    async def delete_bucket_async(self, bucket: str) -> None:
        """Delete bucket, for usage in async applications."""
        await self.delete_by_prefix(bucket=bucket, prefix="")
        async with self._get_async_boto_client() as s3_client:
            await s3_client.delete_bucket(Bucket=bucket)

    def delete_by_prefix(self, bucket: str, prefix: str) -> Optional[asyncio.Task]:
        """Delete all objects with specified prefix in key.

        Parameters
        ----------
        bucket_name: str
            name of bucket you want to delete directory from.
        prefix: str
            path to your objects or dir

            Example::

            |    Directories structure:
            |    ├───dockerfile
            |    ├───a.py
            |    ├───dir1
            |            ├───empty.txt
            |            └───dir2
            |                    ├───new.txt
            |                    ├───x.png
            |                    └───ipsum
            |                            └───y.png
            |    To delete dir1    use prefix="dir1"
            |    To delete dir2    use prefix="dir1/dir2"
            |    To delete new.txt use prefix="dir1/dir2/new.txt"
        """
        try:
            if asyncio.get_event_loop().is_running():
                return asyncio.create_task(self._delete_objects(bucket_name=bucket, prefix=prefix))
            asyncio.get_event_loop().close()
        except Exception:
            pass
        asyncio.run(self._delete_objects(bucket_name=bucket, prefix=prefix))

    def upload(
        self,
        local_path: str,
        bucket: str,
        upload_as_tar: bool = False,
        new_bucket_visibility: VisibilityOptions = VisibilityOptions.PRIVATE,
        verbose: bool = True,
    ) -> Optional[asyncio.Task]:
        """
        Upload directory to bucket.

        Parameters
        ----------
        local_path: str
            path to directory with files you want to upload.
        bucket: str
            name of bucket you want to upload to.
        upload_as_tar: bool = False
            If the option is set to True, the files will be uploaded as a single tar archive. Default: False
        verbose: bool = True
            If the option is set to True and upload_as_tar set to False,
            a progress bar with the number of uploaded files will be displayed.

        Returns
        -------
        Optional[asyncio.Task].
            If the files uploading to the bucket is started inside an asynchronous application,
            the method will schedule the task in the running event loop and
            return instance of asyncio.Task for further process monitoring by the application
        """
        if upload_as_tar:
            self._upload_as_tar(local_path=local_path, bucket=bucket, new_bucket_visibility=new_bucket_visibility)
            return

        # in case of asyncio.run() was called previously in code, it closed event loop
        # and get_event_loop() will raise exception:
        # RuntimeError: There is no current event loop in thread <THREAD_NAME>
        try:
            if asyncio.get_event_loop().is_running():
                return asyncio.create_task(
                    self._async_upload_files(
                        local_path=local_path,
                        bucket=bucket,
                        new_bucket_visibility=new_bucket_visibility,
                        verbose=verbose,
                    )
                )
            asyncio.get_event_loop().close()
        except Exception:
            pass
        asyncio.run(
            self._async_upload_files(
                local_path=local_path, bucket=bucket, new_bucket_visibility=new_bucket_visibility, verbose=verbose
            )
        )

    def set_data(
        self,
        *,
        local_path: str = "s3_data",
        bucket: str,
        untar_data: bool = False,
        remote_paths: Optional[List[str]] = None,
        verbose: bool = True,
    ) -> Union[asyncio.Task, str]:
        """
        Set data.

        :type local_path: string
        :param local_path: Local path to save data to.  Defaults to /s3_data/.

        :type bucket: string
        :param bucket: Bucket containing requested files.

        :type remote_paths: list(string)
        :param remote_paths: List of paths relative to passed bucket.  Each path
            can represent either a single file, or a folder.  If a path represents
            a folder (should end with a slash), then all contents of a folder are recursively downloaded.

        :type verbose: bool
        :param verbose: Whether to disable the entire progressbar wrapper.
        """
        if bucket not in self.list_buckets():
            raise S3BucketNotFoundError(bucket)
        if untar_data:
            return self._download_data_tar(
                local_path=local_path,
                bucket=bucket,
                remote_paths=remote_paths,
            )

        try:
            if asyncio.get_event_loop().is_running():
                return asyncio.create_task(
                    self._download_files(
                        local_path=local_path,
                        bucket=bucket,
                        remote_paths=remote_paths,
                        verbose=verbose,
                    )
                )
            asyncio.get_event_loop().close()
        except Exception:
            pass
        return asyncio.run(
            self._download_files(
                local_path=local_path,
                bucket=bucket,
                remote_paths=remote_paths,
                verbose=verbose,
            )
        )

    def _upload_as_tar(self, local_path: str, bucket: str, new_bucket_visibility: VisibilityOptions) -> None:
        local_path = os.path.normpath(local_path)
        if not os.path.exists(local_path):
            raise FileNotFoundError(f"Path: {local_path} does not exist")

        r, w = os.pipe()

        try:
            thread = threading.Thread(target=_tar_folder, args=(w, local_path), daemon=True)
            thread.start()
        except Exception as err:
            os.close(r)
            os.close(w)
            raise err

        s3_client = self._get_sync_boto_client()
        buckets = self.list_buckets()
        if bucket not in buckets:
            self.create_bucket(name=bucket, visibility=new_bucket_visibility)
        try:
            with open(r, "rb") as fileobj:
                s3_client.upload_fileobj(Fileobj=fileobj, Bucket=bucket, Key=f"{os.path.basename(local_path)}.tar")
        finally:
            thread.join()

    async def _async_upload_files(
        self, local_path: str, bucket: str, new_bucket_visibility: VisibilityOptions, verbose: bool = True
    ):
        local_path = os.path.normpath(local_path)
        if not os.path.exists(local_path):
            raise FileNotFoundError(f"Path: {local_path} does not exist")

        async with self._get_async_boto_client() as s3_client:
            buckets = self.list_buckets()
            if bucket not in buckets:
                self.create_bucket(name=bucket, visibility=new_bucket_visibility)

            upload_paths = get_upload_paths(local_path)
            total_tasks = len(upload_paths)

            with tqdm(
                total=total_tasks,
                disable=not verbose,
                unit="Files",
                unit_scale=True,
                unit_divisor=1024,
            ) as pbar:
                current_task_num = 0
                while current_task_num < total_tasks:
                    tasks: list[asyncio.Task] = []
                    for path in upload_paths[current_task_num : current_task_num + MAX_TASKS_NUMBER]:
                        tasks.append(
                            asyncio.create_task(
                                s3_client.upload_file(Filename=path.local_path, Bucket=bucket, Key=path.storage_path)
                            )
                        )
                    await asyncio.gather(*tasks)
                    pbar.update(len(tasks))
                    current_task_num += MAX_TASKS_NUMBER

    def _download_data_tar(
        self,
        *,
        local_path: str = "s3_data",
        bucket: str,
        remote_paths: Optional[List[str]] = None,
    ) -> str:
        if remote_paths is not None and len(remote_paths) != 1:
            raise RuntimeError(f"Expected one tar object, but {len(remote_paths)} were given.")

        s3_client = self._get_sync_boto_client()
        if remote_paths is not None:
            tar_name = remote_paths[0]
            if not tar_name.endswith(".tar"):
                raise FileIsNotTarError("Expected tar file, but file does not have '.tar' extension")
            return self._download_tar(
                local_path=local_path, s3_client=s3_client, tar_name=remote_paths[0], bucket=bucket
            )

        try:
            # check that there is only one file in bucket
            # If there is more than 1 file, raise exception about ambiguous choice
            paginator = s3_client.get_paginator("list_objects_v2")
            page_iterator = paginator.paginate(Bucket=bucket, Prefix="", PaginationConfig={"MaxItems": 2})
            first_page = next(iter(page_iterator))
            files = first_page.get("Contents", [])
            if len(files) != 1:
                raise AmbiguousFileChoiceError(number_of_files=len(files))
            tar_name: str = files[0]["Key"]
            if not tar_name.endswith(".tar"):
                raise FileIsNotTarError("Expected tar file, but file does not have '.tar' extension")
        except ClientError as err:
            if err.response["Error"]["Code"] == "NoSuchBucket":
                raise S3BucketNotFoundError(bucket=bucket) from None
            else:
                raise

        return self._download_tar(local_path=local_path, s3_client=s3_client, tar_name=tar_name, bucket=bucket)

    def _download_tar(self, local_path: str, tar_name: str, bucket: str, s3_client) -> str:
        r, w = os.pipe()
        with open(r, "rb") as buff:
            try:
                thread = threading.Thread(target=_untar_folder, args=(buff, local_path), daemon=True)
                thread.start()
            except Exception as err:
                os.close(r)
                os.close(w)
                raise err

            try:
                with open(w, "wb") as fileobj:
                    s3_client.download_fileobj(Fileobj=fileobj, Bucket=bucket, Key=tar_name)
            except ClientError as err:
                if err.response["Error"]["Code"] == "404":
                    raise S3ObjectNotFoundError(path=tar_name, bucket=bucket) from None
                raise
            finally:
                thread.join()

            return os.path.join(local_path, tar_name[:-4])

    async def _download_files(
        self,
        *,
        local_path: str = "s3_data",
        bucket: str,
        remote_paths: Optional[List[str]] = None,
        verbose: bool = True,
    ) -> str:
        async def get_objects_number(paginator, remote_path: str = "") -> int:
            total_objects = 0

            page_iterator = paginator.paginate(
                Bucket=bucket,
                Prefix=remote_path,
            )
            async for page in page_iterator:
                if page["KeyCount"] != 0:
                    total_objects += page["KeyCount"]
                    continue
                if remote_path == "":
                    return 0
                raise S3ObjectNotFoundError(path=remote_path, bucket=bucket)

            return total_objects

        remote_paths = remote_paths if remote_paths else [""]
        async with self._get_async_boto_client() as s3_client:
            total_tasks = 0
            try:
                paginator = s3_client.get_paginator("list_objects_v2")
                for remote_path in remote_paths:
                    total_tasks += await get_objects_number(paginator, remote_path)
            except ClientError as err:
                raise RuntimeError("Can not get number of objects in bucket") from err

            with tqdm(
                total=total_tasks,
                unit="Files",
                unit_scale=True,
                unit_divisor=1024,
                disable=not verbose,
            ) as pbar:
                for remote_path in remote_paths:
                    page_iterator = paginator.paginate(
                        Bucket=bucket, Prefix=remote_path, PaginationConfig={"PageSize": MAX_OBJECTS_PER_PAGE}
                    )
                    async for page in page_iterator:
                        tasks: list[Coroutine] = []
                        for obj in page.get("Contents", []):
                            file_path = obj.get("Key")

                            local_dir_path = os.path.join(local_path, posixpath.dirname(file_path))
                            local_file_path = os.path.join(local_path, file_path)
                            tasks.append(
                                self._download_one_object(
                                    bucket=bucket,
                                    key=file_path,
                                    local_dir_path=local_dir_path,
                                    local_file_path=local_file_path,
                                    s3_client=s3_client,
                                )
                            )
                        await asyncio.gather(*tasks)
                        pbar.update(len(tasks))

        return local_path

    async def _delete_objects(self, bucket_name: str, prefix="") -> None:
        if bucket_name not in self.list_buckets():
            raise S3BucketNotFoundError(bucket_name)
        async with self._get_async_boto_client() as s3_client:
            paginator = s3_client.get_paginator("list_objects_v2")
            page_iterator = paginator.paginate(
                Bucket=bucket_name, Prefix=prefix, PaginationConfig={"PageSize": MAX_DELETE_CONCURRENT_REQUESTS}
            )
            async for page in page_iterator:
                tasks: list[Coroutine] = []
                for obj in page.get("Contents", []):
                    object_key = obj.get("Key")
                    tasks.append(self._delete_one_object(bucket=bucket_name, key=object_key, s3_client=s3_client))
                await asyncio.gather(*tasks)

    # arguments to callback are passed like kwargs, so kwargs must be present in signature
    def _add_auth_cookies(self, request: AWSRequest, **kwargs) -> None:  # noqa
        request.headers.add_header("Cookie", self.session._get_cookie_header())

    # arguments to callback are passed like kwargs, so kwargs must be present in signature
    def _update_auth_cookies(self, http_response: AWSResponse, **kwargs) -> None:  # noqa
        cookie_header = http_response.headers.get("set-cookie")
        if cookie_header is None:
            return
        cookies: list[str] = cookie_header.split("; ")
        for cookie in cookies:
            if "kc-access" not in cookie:
                continue
            _, new_access_token = cookie.split("=", maxsplit=1)
            self.session.cookies["kc-access"] = new_access_token
            break

    async def _download_one_object(self, bucket: str, key: str, local_dir_path: str, local_file_path: str, s3_client):
        if not await aos.path.exists(local_dir_path):
            await aos.makedirs(local_dir_path, exist_ok=True)
        await s3_client.download_file(Bucket=bucket, Key=key, Filename=local_file_path)

    async def _delete_one_object(self, bucket: str, key: str, s3_client, retries: int = 3):
        # do retries if there are some network troubles, cause bucket and dir deleting depends on it,
        # if object was deleted earlier, s3_client.delete_object has no effect
        while retries > 0:
            try:
                await s3_client.delete_object(Bucket=bucket, Key=key)
                break
            except Exception:
                retries -= 1
                if retries == 0:
                    raise
                await asyncio.sleep(random.randint(1, 9) * 1e-4)

    def _get_sync_boto_client(self):
        s3_client = boto3.client(
            service_name="s3",
            use_ssl=True,
            endpoint_url=posixpath.join(self.default_url, "s3/"),
            aws_access_key_id=self.default_access_key_id,
            aws_secret_access_key=self.default_secret_access_key,
        )
        event_system = s3_client.meta.events
        event_system.register("before-sign.s3.*", self._add_auth_cookies)
        event_system.register("after-call.s3.*", self._update_auth_cookies)
        return s3_client

    @asynccontextmanager
    async def _get_async_boto_client(self) -> AsyncGenerator[Any, None]:
        session = aioboto3.Session(
            aws_access_key_id=self.default_access_key_id, aws_secret_access_key=self.default_secret_access_key
        )
        async with session.client(
            "s3", use_ssl=True, endpoint_url=posixpath.join(self.default_url, "s3/")
        ) as s3_client:
            event_system = s3_client.meta.events
            event_system.register("before-sign.s3.*", self._add_auth_cookies)
            event_system.register("after-call.s3.*", self._update_auth_cookies)
            yield s3_client



"""Executor template for custom executor."""
from typing import List

from ML_management.executor.base_executor import BaseExecutor
from ML_management.executor.patterns import OneModelPattern
from ML_management.executor.upload_model_mode import UploadModelMode
from ML_management.model.model_type_to_methods_map import ModelMethodName


class JobExecutorPattern(BaseExecutor):
    """DEPRECATED.

    Exists only for backward compatibility.
    Instead use BaseExecutor from ML_management.executor.base_executor.
    """

    def __init__(self, desired_model_methods: List[ModelMethodName], upload_model_mode: UploadModelMode) -> None:
        super().__init__(
            executor_models_pattern=OneModelPattern(
                upload_model_mode=upload_model_mode, desired_model_methods=desired_model_methods
            ),
        )



import unittest
from typing import Dict, List, Optional, Union

from ML_management.mlmanagement.jsonschema_exceptions import (
    DictKeysMustBeStringsError,
    FunctionContainsVarArgsError,
    FunctionContainsVarKwArgsError,
    InvalidStructureAnnotationError,
    NoAnnotationError,
    UnsupportedTypeError,
)
from ML_management.mlmanagement.jsonschema_inference import infer_jsonschema


class TestJsonschemaInference(unittest.TestCase):
    def test_jsonschema_generation(self):
        """Tests automatic jsonschema generation from functions signature."""

        def one_1(
            *,
            a: int,
            b: Optional[int],
            c: bool,
            d: Union[int, str],
            e: Dict[str, List[int]],
            f: List[float],
            g: float = 3.23,
        ):
            pass

        result = infer_jsonschema(one_1, "get_object")
        result["required"].sort()
        expected_result = {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "type": "object",
            "properties": {
                "a": {"type": "integer"},
                "b": {"type": ["integer", "null"]},
                "c": {"type": "boolean"},
                "d": {"anyOf": [{"type": "integer"}, {"type": "string"}]},
                "e": {
                    "type": "object",
                    "additionalProperties": {
                        "type": "array",
                        "items": {"type": "integer"},
                    },
                },
                "f": {"type": "array", "items": {"type": "number"}},
                "g": {"type": "number", "default": 3.23},
            },
            "required": ["a", "c", "d", "e", "f"],
            "additionalProperties": False,
        }
        self.assertDictEqual(result, expected_result)

        def one_2(
            a: int,
            b: Optional[int],
            c: bool,
            d: Union[int, str],
            *,
            e: Dict[str, List[int]],
            f: List[float],
            g: float = 3.23,
        ):
            pass

        result = infer_jsonschema(one_2, "get_object")
        result["required"].sort()
        self.assertDictEqual(result, expected_result)

        def one_3(
            a: int,
            b: Optional[int],
            c: bool,
            d: Union[int, str],
            e: Dict[str, List[int]],
            f: List[float],
            g: float = 3.23,
        ):
            pass

        result = infer_jsonschema(one_3, "get_object")
        result["required"].sort()
        self.assertDictEqual(result, expected_result)

        def two(a):
            pass

        self.assertRaises(NoAnnotationError, infer_jsonschema, two, "get_object")

        def three(a: complex):
            pass

        self.assertRaises(UnsupportedTypeError, infer_jsonschema, three, "get_object")

        def four(a: Dict[int, str]):
            pass

        self.assertRaises(DictKeysMustBeStringsError, infer_jsonschema, four, "get_object")

        class SomeType:
            def __init__(self) -> None:
                pass

        def five(a: SomeType):
            pass

        self.assertRaises(UnsupportedTypeError, infer_jsonschema, five, "get_object")

        def six(a: int, **kwargs):
            pass

        self.assertRaises(FunctionContainsVarKwArgsError, infer_jsonschema, six, "get_object")

        def seven(a: int, *args):
            pass

        self.assertRaises(FunctionContainsVarArgsError, infer_jsonschema, seven, "get_object")

        def eight(a: List):
            pass

        self.assertRaises(InvalidStructureAnnotationError, infer_jsonschema, eight, "get_object")


if __name__ == "__main__":
    unittest.main()



"""Module for automatic jsonschema inference."""
import warnings
from copy import copy
from inspect import formatannotation, getfullargspec
from typing import Generic, TypeVar

from jsonschema import SchemaError
from jsonschema.validators import Draft4Validator

from ML_management.mlmanagement.jsonschema_exceptions import (
    DictKeysMustBeStringsError,
    FunctionContainsVarArgsError,
    FunctionContainsVarKwArgsError,
    InvalidSchemaError,
    InvalidStructureAnnotationError,
    NoAnnotationError,
    UnsupportedTypeError,
)

JSON_SCHEMA_DRAFT = "http://json-schema.org/draft-04/schema#"


type_map = {
    "bool": {"type": "boolean"},
    "int": {"type": "integer"},
    "float": {"type": "number"},
    "str": {"type": "string"},
}


T = TypeVar("T")


class SkipJsonSchema(Generic[T]):
    """Annotation wrapper for system parameters to skip them in JSON schema."""

    def __class_getitem__(cls, _type):
        return cls()


def __get_or_raise(type_name, from_optional: bool = False):
    if type_name in type_map:
        field = copy(type_map[type_name])
        if from_optional:
            field["type"] = [field["type"], "null"]
        return field
    else:
        raise UnsupportedTypeError(annotation=type_name, supported_types=list(type_map.keys()))


def __is_optional(annotation):
    repr_annotation = repr(annotation)
    if repr_annotation.startswith("typing.Optional"):
        return True
    elif repr_annotation.startswith("typing.Union") and repr_annotation.strip("]").endswith("NoneType"):
        # python 3.8 casts Optional to Union[arg, None]
        return True
    else:
        return False


def __get_json_schema_from_annotation(annotation, from_optional: bool = False):
    if annotation.__module__ == "typing":
        if __is_optional(annotation):
            return __get_json_schema_from_annotation(
                annotation.__args__[0], from_optional=True
            )  # Optional[int] translates to type: [integer, null], and the field is not required
        else:
            formatted_annotation = formatannotation(annotation)

            # this is the only way to reliably get annotation name (e.g. "List")
            annotation_name = formatted_annotation.partition("[")[0]
            if not hasattr(annotation, "__args__"):
                raise InvalidStructureAnnotationError(annotation=formatted_annotation)
            if annotation_name == "List":
                return {
                    "type": "array" if not from_optional else ["array", "null"],
                    "items": __get_json_schema_from_annotation(annotation.__args__[0]),
                }
            elif annotation_name == "Dict":
                key_annotation = annotation.__args__[0]
                if hasattr(key_annotation, "__name__"):
                    if key_annotation.__name__ != "str":
                        raise DictKeysMustBeStringsError(annotation=formatted_annotation)
                else:
                    raise DictKeysMustBeStringsError(annotation=formatted_annotation)
                return {
                    "type": "object" if not from_optional else ["object", "null"],
                    "additionalProperties": __get_json_schema_from_annotation(annotation.__args__[1]),
                }
            elif annotation_name == "Union":
                union_types = [__get_json_schema_from_annotation(ann) for ann in annotation.__args__]
                marked_as_null = False
                for type_ in union_types:
                    if type_ == {"type": "null"}:
                        marked_as_null = True
                        break
                if not marked_as_null and from_optional:
                    union_types.extend([{"type": "null"}])
                return {"anyOf": union_types}
            else:
                raise UnsupportedTypeError(
                    annotation=formatannotation(annotation),
                    supported_types=list(type_map.keys()),
                )

    elif annotation.__module__ == "builtins":
        if hasattr(annotation, "__name__"):
            return __get_or_raise(annotation.__name__, from_optional=from_optional)
    else:
        raise UnsupportedTypeError(
            annotation=formatannotation(annotation),
            supported_types=list(type_map.keys()),
        )


def get_function_args(spec, func):
    """Get all arguments of function."""
    # FullArgSpec(args, varargs, varkw, defaults, kwonlyargs, kwonlydefaults, annotations)
    pos_args_defaults = {}
    if spec.defaults:
        for arg, default in zip(spec.args[-len(spec.defaults) :], spec.defaults):
            pos_args_defaults[arg] = default
    if spec.varargs:
        raise FunctionContainsVarArgsError(function_name=func.__name__)
    if spec.varkw:
        raise FunctionContainsVarKwArgsError(function_name=func.__name__)

    # annotations are not inferred from default yet, but i can do it myself from kwonlydefaults
    pos_args = spec.args if spec.args else []
    kw_only_args = spec.kwonlyargs if spec.kwonlyargs else []
    all_args = pos_args + kw_only_args
    kw_only_args_defaults = spec.kwonlydefaults if spec.kwonlydefaults else {}
    all_defaults = {**pos_args_defaults, **kw_only_args_defaults}

    return all_args, all_defaults


def infer_jsonschema(func, get_object_func):
    """Infer jsonschema from callable by its signature."""
    from ML_management.model.model_type_to_methods_map import ModelMethodName  # circular import

    spec = getfullargspec(func)
    all_args, all_defaults = get_function_args(spec, func)
    schema = {
        "$schema": JSON_SCHEMA_DRAFT,
        "type": "object",
        "properties": {},
        "additionalProperties": False,
    }
    if "self" in all_args:
        all_args.remove("self")

    if func.__name__ == ModelMethodName.init.value and get_object_func:
        spec_get_object = getfullargspec(get_object_func)
        all_args_get_object, _ = get_function_args(spec_get_object, get_object_func)
        if not set(all_args_get_object) == set(all_args):
            warnings.warn(
                """get_object function arguments do not match __init__ arguments.
                You won't be able to pass none default arguments into __init__ function."""
            )
            all_args = []
            all_defaults = []

    required = []
    for arg in all_args:
        if arg not in spec.annotations:
            raise NoAnnotationError(arg=arg)

        # skip system parameters
        if isinstance(spec.annotations[arg], SkipJsonSchema):
            continue

        schema["properties"][arg] = __get_json_schema_from_annotation(spec.annotations[arg])
        if arg in all_defaults:
            schema["properties"][arg]["default"] = all_defaults[arg]
        if arg not in all_defaults and not __is_optional(spec.annotations[arg]):
            required.append(arg)

    if required:
        schema["required"] = required
    try:
        Draft4Validator.check_schema(schema)
    except SchemaError as err:
        raise InvalidSchemaError(schema=schema, original_message=str(err)) from None

    return schema



"""Class inherits from base class of Model with a specific method for torch target layer."""
from abc import ABC, abstractmethod

from ML_management.model.patterns import model_pattern


class TargetLayer(model_pattern.Model, ABC):
    """Implementation of model with target_layer function."""

    def __init__(self):
        super().__init__()

    @abstractmethod
    def get_target_layer(self, **kwargs):
        """Define target_layer."""
        raise NotImplementedError



from sgqlc.operation import Operation

from ML_management.sdk import schema
from ML_management.sdk.schema import Experiment
from ML_management.sdk.sdk import send_graphql_request


def set_experiment_description(experiment_name: str, description: str) -> Experiment:
    """
    Set experiment description.

    Parameters
    ----------
    experiment_name: str
        Name of an experiment.
    description: str
        Description of an experiment.

    Returns
    -------
    Experiment
        Instance of a experiment with meta information.
    """
    op = Operation(schema.Mutation)
    set_description = op.set_experiment_description(name=experiment_name, description=description)
    set_description.name()
    set_description.description()

    experiment = send_graphql_request(op=op)
    return experiment["setExperimentDescription"]


def set_experiment_tag(experiment_name: str, key: str, value: str) -> Experiment:
    """
    Set experiment tag.

    Parameters
    ----------
    experiment_name: str
        Name of an experiment.
    key: str
        Key of a tag.
    value: str
        Value of a tag.

    Returns
    -------
    Experiment
        Instance of an experiment with meta information.
    """
    op = Operation(schema.Mutation)
    set_experiment_tags = op.set_experiment_tag(name=experiment_name, key=key, value=value)
    set_experiment_tags.name()
    set_experiment_tags.tags()

    experiment = send_graphql_request(op=op)
    return experiment["setExperimentTag"]


def delete_experiment_tag(experiment_name: str, key: str) -> Experiment:
    """
    Delete experiment tag.

    Parameters
    ----------
    experiment_name: str
        Name of an experiment.
    key: str
        Key of a tag to delete.

    Returns
    -------
    Experiment
        Instance of an experiment with meta information.
    """
    op = Operation(schema.Mutation)
    set_experiment_tags = op.delete_experiment_tag(name=experiment_name, key=key)
    set_experiment_tags.name()
    set_experiment_tags.tags()

    experiment = send_graphql_request(op=op)
    return experiment["deleteExperimentTag"]


def get_experiment_by_name(experiment_name: str) -> Experiment:
    """
    Get experiment by it's name.

    Parameters
    ----------
    experiment_name: str
        Name of an experiment.

    Returns
    -------
    Experiment
        Instance of an experiment with meta information.
    """
    op = Operation(schema.Query)
    experiment_from_name = op.experiment_from_name(name=experiment_name)
    experiment_from_name.name()
    experiment_from_name.tags()
    experiment_from_name.description()
    experiment_from_name.experiment_id()

    experiment = send_graphql_request(op=op)
    return experiment["experimentFromName"]


def get_experiment_by_id(experiment_id: str) -> Experiment:
    """
    Get experiment by it's id.

    Parameters
    ----------
    experiment_id: str
        Id of an experiment.

    Returns
    -------
    Experiment
        Instance of an experiment with meta information.
    """
    op = Operation(schema.Query)
    experiment_from_name = op.experiment_from_id(experiment_id=experiment_id)
    experiment_from_name.name()
    experiment_from_name.tags()
    experiment_from_name.description()
    experiment_from_name.experiment_id()

    experiment = send_graphql_request(op=op)
    return experiment["experimentFromId"]



"""Module with functionality for detecting used local modules."""
import ast
import importlib
import inspect
import site
import sys
import sysconfig
from typing import Set, Tuple


class ModuleFinder:
    """Class with functions for modules detection."""

    def _get_imported_modules(self, name: str, node_iter: ast.NodeVisitor) -> set:
        self.modules = set()
        try:
            filename = inspect.getfile(sys.modules[name])
        except Exception:
            return self.modules
        with open(filename) as f:
            node_iter.visit(ast.parse(f.read()))
        for module_name in self.modules.copy():
            try:
                module = sys.modules[module_name]
            except KeyError:
                module = importlib.import_module(module_name)  # raise ModuleNotFound if module does not present.
            if self.check_module_is_not_local(module_name, module):
                self.modules.remove(module_name)

        return self.modules

    def _visit_import(self, node):
        for name in node.names:
            self.modules.add(name.name)

    def _visit_import_from(self, node):
        if node.module is not None and node.level == 0:
            self.modules.add(node.module)

    def _find_submodules(self, name: str) -> set:
        node_iter = ast.NodeVisitor()
        node_iter.visit_Import = self._visit_import
        node_iter.visit_ImportFrom = self._visit_import_from
        module_set = self._get_imported_modules(name, node_iter)

        for module in module_set:
            if module != name:
                module_set = module_set.union(self._find_submodules(module))

        return module_set

    def find_root_submodules(self, root_name: str = "__main__") -> Set[str]:
        """Find all local submodules used by the root module."""
        if hasattr(sys.modules[root_name], "__file__"):
            submodules: set[str] = self._find_submodules(root_name)
            submodules.add(root_name)
            return submodules

        submodules: set[str] = set()
        for obj_name, obj in inspect.getmembers(sys.modules[root_name]):
            try:
                module_name, module_obj = self.get_module_name_obj_for_classes_funcs_modules(obj_name, obj)
                if self.check_module_is_not_local(module_name, module_obj):
                    continue
                submodules = submodules | {module_name} | self._find_submodules(module_name)
            except TypeError:
                continue
        return (submodules | {root_name}) - {"__builtin__", "__builtins__"}

    @staticmethod
    def get_module_name_obj_for_classes_funcs_modules(obj_name, obj) -> Tuple[str, object]:
        """Get module name and module object for class, function or module."""
        if inspect.isclass(obj) or inspect.isfunction(obj):
            return obj.__module__, sys.modules[obj.__module__]
        if inspect.ismodule(obj):
            return obj_name, obj
        raise TypeError("Object is not either function, class or module")

    @staticmethod
    def import_modules_by_name(names: Set[str]) -> None:
        """Import modules by it's names."""
        for name in names:
            importlib.import_module(name=name)

    @staticmethod
    def check_module_is_not_local(module_name, module_obj):
        """Check if module is local file or not."""
        paths = sysconfig.get_paths()
        return (
            paths["stdlib"] in str(module_obj)
            or paths["platstdlib"] in str(module_obj)
            or paths["platlib"] in str(module_obj)
            or paths["purelib"] in str(module_obj)
            or module_name in sys.builtin_module_names
            or "ML_management" in str(module_obj)
            or (site.USER_SITE) in str(module_obj)
        )

    @staticmethod
    def get_my_caller_module_name() -> str:
        """Return the name of the module in which the function was called or None."""
        module = inspect.getmodule(inspect.stack()[2][0])
        # for example for code, that is result of cloudpickle.load
        # and does not have module.
        if module is None:
            return "__main__"
        return module.__name__



import importlib
import os
import os.path
import subprocess
import sys
import tarfile
import tempfile
import threading
import traceback
from typing import Optional

from ML_management.mlmanagement.base_exceptions import MLMClientError
from ML_management.mlmanagement.log_api import _raise_error
from ML_management.mlmanagement.model_type import ModelType
from ML_management.mlmanagement.session import AuthSession
from ML_management.mlmanagement.variables import get_log_service_url
from mlflow.models import Model
from mlflow.pyfunc import (
    CODE,
    DATA,
    FLAVOR_NAME,
    MAIN,
    MLMODEL_FILE_NAME,
    PY_VERSION,
    RESOURCE_DOES_NOT_EXIST,
    MlflowException,
    PyFuncModel,
    _add_code_from_conf_to_system_path,
    _download_artifact_from_uri,
    _warn_dependency_requirement_mismatches,
    _warn_potentially_incompatible_py_version_if_necessary,
)


def download_artifacts_by_name_version(
    name: str, version: int, model_type: ModelType, path: str, dst_path: Optional[str] = None
) -> str:
    """Download an artifact by name and version to a local directory, and return a local path for it."""
    url = get_log_service_url("download_artifacts_by_name_version")
    params = {
        "path": os.path.normpath(path) if path else path,
        "name": name,
        "version": version,
        "model_type": model_type.value,
    }
    return _request_download_artifacts(url, params, dst_path)


def download_job_artifacts(job_id: str, path: str = "", dst_path: Optional[str] = None) -> str:
    """Download an artifact file or directory from a job to a local directory, and return a local path for it."""
    url = get_log_service_url("download_job_artifacts")
    params = {"path": os.path.normpath(path) if path else path, "job_name": job_id}
    return _request_download_artifacts(url, params, dst_path)


def _load_model_type(
    name: str,
    version: int,
    model_type: ModelType,
    unwrap: bool = True,
    install_requirements: bool = False,
    dst_path: Optional[str] = None,
    kwargs_for_init=None,
):
    """Load model from local path."""
    local_path = download_artifacts_by_name_version(
        name=name, version=version, model_type=model_type, path="", dst_path=dst_path
    )
    if install_requirements:
        _set_model_version_requirements(local_path)
    model_load_dict = {"model_path": local_path, "kwargs_for_init": kwargs_for_init}
    loaded_model = _load_model_mflow(model_data=model_load_dict, suppress_warnings=True)
    if unwrap:
        artifacts_path = loaded_model._model_impl.context._artifacts
        loaded_model = loaded_model.unwrap_python_model()
        loaded_model.artifacts = artifacts_path
    return loaded_model


def load_dataset(
    name: str,
    version: int,
    install_requirements: bool = False,
    unwrap: bool = True,
    dst_path: Optional[str] = None,
    kwargs_for_init: Optional[dict] = None,
):
    """Download all model's files for loading model locally.

    Parameters
    ==========
    name: str
        Name of the dataset.
    version: int
        Version of the dataset.
    install_requirements: bool = False
        Whether to install dataset requirements. Default: False.
    unwrap: bool = True
        Whether to unwrap dataset. Default: True.
    dst_path: Optional[str]: None
        Destination path. Default: None.
    kwargs_for_init: Optional[dict]: None
        Kwargs for __init__ function of dataset loader when it would be loaded.
    Returns
    =======
    DatasetLoaderPattern
        The object of the dataset to use.
    """
    return _load_model_type(
        name, version, ModelType.DATASET_LOADER, unwrap, install_requirements, dst_path, kwargs_for_init
    )


def _set_model_version_requirements(local_path) -> None:
    """Installing requirements of the model locally."""
    with open(os.path.join(local_path, "requirements.txt")) as req:
        requirements = list(
            filter(lambda x: "ml-management" not in x.lower() and "mlflow" not in x.lower(), req.read().split("\n"))
        )
    try:
        if requirements:
            subprocess.check_call(
                [sys.executable, "-m", "pip", "install", "--no-cache-dir", "--default-timeout=100", *requirements]
            )
    except Exception:
        print(traceback.format_exc())


def load_model(
    name: str,
    version: int,
    install_requirements: bool = False,
    unwrap: bool = True,
    dst_path: Optional[str] = None,
    kwargs_for_init=None,
):
    """Download all model's files for loading model locally.

    Parameters
    ==========
    name: str
        Name of the model.
    version: int
        Version of the model.
    install_requirements: bool = False
        Whether to install model requirements. Default: False.
    unwrap: bool = True
        Whether to unwrap model. Default: True.
    dst_path: Optional[str]: None
        Destination path. Default: None.
    kwargs_for_init: Optional[dict]: None
        Kwargs for __init__ function of model when it would be loaded.
    Returns
    =======
    Model
        The object of the model to use.
    """
    return _load_model_type(name, version, ModelType.MODEL, unwrap, install_requirements, dst_path, kwargs_for_init)


def load_executor(
    name: str,
    version: int,
    install_requirements: bool = False,
    unwrap: bool = True,
    dst_path: Optional[str] = None,
):
    """Download all model's files for loading model locally.

    Parameters
    ==========
    name: str
        Name of the executor.
    version: int
        Version of the executor.
    install_requirements: bool = False
        Whether to install executor requirements. Default: False.
    unwrap: bool = True
        Whether to unwrap executor. Default: True.
    dst_path: Optional[str]: None
        Destination path. Default: None.
    Returns
    =======
    BaseExecutor
        The object of the executor to use.
    """
    return _load_model_type(name, version, ModelType.EXECUTOR, unwrap, install_requirements, dst_path)


def _untar_folder(buff, to_folder):
    try:
        with tarfile.open(mode="r|", fileobj=buff) as tar:
            tar.extractall(to_folder)
    except Exception as err:
        raise MLMClientError("Some error during untar the content.") from err


def _request_download_artifacts(url, params: dict, dst_path: Optional[str] = None):
    path = params["path"]
    with AuthSession().get(url=url, params=params, stream=True) as response:
        _raise_error(response)
        untar = response.headers.get("untar") == "True"
        if dst_path is None:
            dst_path = tempfile.mkdtemp()
        dst_path = os.path.abspath(os.path.normpath(dst_path))
        local_path = os.path.normpath(os.path.join(dst_path, os.path.normpath(path)))
        if untar:
            r, w = os.pipe()
            with open(r, "rb") as buff:
                try:
                    thread = threading.Thread(target=_untar_folder, args=(buff, local_path))
                    thread.start()
                except Exception as err:
                    os.close(r)
                    os.close(w)
                    raise err

                with open(w, "wb") as wfd:
                    for chunk in response.iter_raw():
                        wfd.write(chunk)
                thread.join()
                return local_path
        else:
            dirs = os.path.dirname(local_path)
            if not os.path.exists(dirs):
                os.makedirs(dirs)
            with open(local_path, "wb") as f:
                for chunk in response.iter_bytes():
                    f.write(chunk)
            return local_path


def _load_model_mflow(
    model_data: dict,
    suppress_warnings: bool = False,
    dst_path: Optional[str] = None,
) -> PyFuncModel:
    """Load model function from mlflow with custom arguments for our loader of a model."""
    # Mflow checks type of arguments so we need to use json as str.
    model_args = model_data
    model_uri = model_args["model_path"]
    local_path = _download_artifact_from_uri(artifact_uri=model_uri, output_path=dst_path)

    if not suppress_warnings:
        _warn_dependency_requirement_mismatches(local_path)

    model_meta = Model.load(os.path.join(local_path, MLMODEL_FILE_NAME))

    conf = model_meta.flavors.get(FLAVOR_NAME)
    if conf is None:
        raise MlflowException(
            f'Model does not have the "{FLAVOR_NAME}" flavor',
            RESOURCE_DOES_NOT_EXIST,
        )
    model_py_version = conf.get(PY_VERSION)
    if not suppress_warnings:
        _warn_potentially_incompatible_py_version_if_necessary(model_py_version=model_py_version)

    _add_code_from_conf_to_system_path(local_path, conf, code_key=CODE)
    data_path = os.path.join(local_path, conf[DATA]) if (DATA in conf) else local_path
    model_args["model_path"] = data_path
    if importlib.import_module(conf[MAIN]).__name__ == "ML_management.loader.loader":
        model_data = model_args
    else:
        model_data = data_path
    model_impl = importlib.import_module(conf[MAIN])._load_pyfunc(model_data)
    predict_fn = conf.get("predict_fn", "predict")
    return PyFuncModel(model_meta=model_meta, model_impl=model_impl, predict_fn=predict_fn)



"""SDK for client library."""
import json
import locale
import posixpath
from typing import Any, Dict, List

import httpx
import pandas as pd
from jsf import JSF
from sgqlc.operation import Operation

from ML_management.mlmanagement import get_server_url
from ML_management.mlmanagement.session import AuthSession


def _to_datetime(df: pd.DataFrame, column_names: List[str]) -> pd.DataFrame:
    """
    Convert df's columns to datetime.

    Parameters
    ----------
    df: pd.DataFrame
        pd.DataFrame in which the columns will be converted.
    column_names: List[str]
        Column names to be converted.

    Returns
    -------
    pd.DataFrame
        Pandas dataframe with converted columns.
    """
    for column_name in column_names:
        df[column_name] = pd.to_datetime(df[column_name], unit="s")

    return df


def send_graphql_request(op: Operation, json_response: bool = True) -> Any:
    """Send request to server and process the response."""
    json_data = AuthSession().sgqlc_request(op)

    if "data" not in json_data or json_data["data"] is None:
        server_url = get_server_url()
        try:
            if locale.getdefaultlocale()[0][:2] == "ru":
                url_base = posixpath.join(server_url, "locales/ru/ru.json")
            else:
                url_base = posixpath.join(server_url, "locales/en/en.json")
        except Exception:
            # if there is no locale file use english by default.
            url_base = posixpath.join(server_url, "locales/en/en.json")
        translation = httpx.get(url_base).json()

        error_message = json_data["errors"][0]["message"].split(",")[0]

        try:
            message_type, message_value = error_message.split(".")
        except Exception:
            raise Exception(error_message) from None

        if message_type not in translation:
            raise Exception(message_type)

        formatted_translated_message = translation[message_type][message_value]
        if (
            ("extensions" in json_data["errors"][0])
            and ("params" in json_data["errors"][0]["extensions"][error_message])
            and (json_data["errors"][0]["extensions"][error_message]["params"] is not None)
        ):
            raise Exception(
                formatted_translated_message.format().format(
                    **json_data["errors"][0]["extensions"][error_message]["params"]
                )
            )
        raise Exception(formatted_translated_message)

    if json_response:
        return json_data["data"]
    else:
        return op + json_data


def _generate_fake_schema(json_schema: dict) -> dict:
    if "required" not in json_schema.keys():
        return {}

    required_properties = {key: json_schema["properties"][key] for key in json_schema["required"]}
    json_schema["properties"] = required_properties

    faker = JSF(json_schema)
    fake_json = faker.generate()
    return fake_json


def _print_params_by_schema(json_schema: Dict, schema_type: str) -> None:
    """Print entity JSON Schema and example with required params."""
    properties_and_required_dict = {key: json_schema[key] for key in ("properties", "required") if key in json_schema}

    json_formatted_str = json.dumps(properties_and_required_dict, indent=2)

    print(f"{schema_type} json-schema:")

    print(json_formatted_str)

    print(f"{schema_type} parameters example:")

    fake_json = _generate_fake_schema(json_schema)

    print(fake_json)



"""S3 Collector for downloading files and folders."""
import asyncio
from typing import List, Optional, Union

from ML_management.collectors.collector_pattern import CollectorPattern
from ML_management.s3 import S3Manager


class S3Collector(CollectorPattern):
    """Collector for S3 paths using ML_management.s3.manager.S3Manager."""

    @staticmethod
    def get_json_schema():
        """Return json schema."""
        return {
            "type": "object",
            "properties": {
                "bucket": {"type": "string"},
                "untar_data": {"type": "boolean"},
                "remote_paths": {"type": "array", "items": {"type": "string"}},
            },
            "required": ["bucket"],
            "additionalProperties": False,
        }

    def set_data(
        self,
        *,
        local_path: str = "s3_data",
        bucket: str,
        untar_data: bool = False,
        remote_paths: Optional[List[str]] = None,
        verbose: bool = True,
    ) -> Union[asyncio.Task, str]:
        """
        Set data.

        :type local_path: string
        :param local_path: Local path to save data to.  Defaults to "s3_data".

        :type bucket: string
        :param bucket: Bucket containing requested files.

        :type remote_paths: list(string)
        :param remote_paths: List of paths relative to passed bucket.  Each path
            can represent either a single file, or a folder.  If a path represents
            a folder (should end with a slash), then all contents of a folder are recursively downloaded.

        :type verbose: bool
        :param verbose: Whether to disable the entire progressbar wrapper.
        """
        return S3Manager().set_data(
            local_path=local_path, bucket=bucket, untar_data=untar_data, remote_paths=remote_paths, verbose=verbose
        )



import json
import posixpath
import sys
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple, Union
from urllib.parse import quote

import numpy as np
import pandas as pd
import websocket
from matplotlib import pyplot as plt
from sgqlc.operation import Operation

from ML_management.mlmanagement import get_server_url
from ML_management.mlmanagement.session import AuthSession
from ML_management.mlmanagement.variables import get_server_websocket_url
from ML_management.sdk import schema
from ML_management.sdk.model import get_latest_model_version
from ML_management.sdk.parameters import ArbitraryDatasetLoaders, ArbitraryModels, SingleDatasetLoader, SingleModel
from ML_management.sdk.schema import ExecutionJob, Param
from ML_management.sdk.sdk import send_graphql_request


def job_by_name(name: str) -> ExecutionJob:
    """
    Return Job object by name.

    Parameters
    ----------
    name: str
        Name of the job.

    Returns
    -------
    Job
        Instance of the Job class.
    """
    op = Operation(schema.Query)

    base_query = op.job_from_name(name=name)
    base_query.name()
    base_query.periodic_type()
    base_query.status()
    base_query.registration_timestamp()
    base_query.start_timestamp()
    base_query.end_timestamp()
    base_query.build_job()
    base_query.exception()

    job = send_graphql_request(op, json_response=False)

    return job.job_from_name


def job_metric_by_name(name: str) -> pd.DataFrame:
    """
    Job's most recent logged metrics.

    Parameters
    ----------
    name: str
        Name of the job.

    Returns
    -------
    pd.DataFrame
        Pandas dataframe with latest metrics.
    """
    op = Operation(schema.Query)

    op.job_from_name(name=name).latest_metrics()
    json_data = send_graphql_request(op)

    json_data = json_data["jobFromName"]["latestMetrics"]

    return pd.DataFrame([json_data])


def available_metrics(job_name: str) -> List[str]:
    """
    List logged types of logged metrics in given job.

    Parameters
    ----------
    job_name: str
        Name of the job.

    Returns
    -------
    List[str]
        List with names of metrics.
    """
    op = Operation(schema.Query)
    base = op.job_from_name(name=job_name)
    base.available_metrics()
    job = send_graphql_request(op, False)

    return job.job_from_name.available_metrics


def list_params_job(job_name: str) -> List[Param]:
    """
    List logged params in given job.

    Parameters
    ----------
    job_name: str
        Name of the job.

    Returns
    -------
    List[Param]
        List of params.
    """
    op = Operation(schema.Query)
    base = op.job_from_name(name=job_name)
    base.list_params()
    job = send_graphql_request(op, False)

    return job.job_from_name.list_params


def metric_history(
    job_name: str, metric_name: str, x_axis: str = "step", make_graph: bool = True, **kwargs
) -> Tuple[List[float], List[float]]:
    """
    Lists history of given metric in given job and plots graph of metric's change over time.

    Parameters
    ----------
    job_name: str
        Name of the job.
    metric_name: str
        Name of the metric.
    x_axis: str
        Witch value use as x axis for graph or as return value. Possible options: "time", "step"
    make_graph: bool
        If to build graph of metric over time.
    kwargs: Any
        Any additional key arguments for plt.plot.

    Returns
    -------
    Tuple(List[float], List[float])
        First list with provided x axis of logged metric.
        Second list with values of metrics.
        Also plots graph using provided x axis and metric's values as y axis.
    """
    op = Operation(schema.Query)
    base = op.job_from_name(name=job_name)
    base.metric_history(metric=metric_name)
    job = send_graphql_request(op, False)
    metrics = job.job_from_name.metric_history
    metric_values = [metric.value for metric in metrics]
    x_values = []
    if x_axis != "step" and x_axis != "time":
        raise ValueError("x_axis value must be step or time")
    if x_axis == "step":
        x_values = [metric.step for metric in metrics]
        if np.nonzero(x_values)[0].size == 0:
            metrics = sorted(metrics, key=lambda x: x.timestamp)
            x_values = [step[0] for step in enumerate(metrics)]
        else:
            metrics = sorted(metrics, key=lambda x: x.step)
    elif x_axis == "time":
        metrics = sorted(metrics, key=lambda x: x.timestamp)
        start_timestamp = metrics[0].timestamp
        x_values = [(metric.timestamp - start_timestamp) / 1000 for metric in metrics]

    if make_graph:
        plt.grid()
        plt.xlabel("seconds" if x_axis == "time" else "step")
        plt.ylabel(metric_name)
        plt.plot(x_values, metric_values, marker=".", **kwargs)

    return x_values, metric_values


def get_logs(job_name: str, stream: bool = True, file_name: Optional[str] = None) -> None:
    """
    Stream logs of the execution job by job name.

    Parameters
    ----------
    job_name: str
        Name of the execution job whose logs we want to view.
    stream: bool = True
        Stream logs or dump all available at the moment.
    file_name: Optional[str] = None
        Name of the file where to save logs. Default: None. If None prints logs to the output.
    """
    _get_logs(job_type=JobType.execution, stream=stream, file_name=file_name, params={"job_id": job_name})


def cancel_job(job_name: str) -> bool:
    """
    Cancel running or planned execution job.

    Parameters
    ----------
    job_name: str
        Name of the job to cancel.
    """
    op = Operation(schema.Mutation)
    op.cancel_job(job_name=job_name)
    return send_graphql_request(op)["cancelJob"]


class JobType(str, Enum):
    """Enum job type."""

    build = "build"
    execution = "execution"
    venv = "venv"


def get_build_logs(
    model_name: str, model_version: Optional[int] = None, stream: bool = True, file_name: Optional[str] = None
) -> None:
    """
    Stream logs of the build job by model name and version.

    Parameters
    ----------
    model_name: str
        The name of the model whose docker image creation logs you want to view.
    model_version: Optional[int] = None
        The version of the model whose docker image creation logs you want to view.
        Default: None, "latest" version is used.
    stream: bool = True
        Stream logs or dump all available at the moment.
    file_name: Optional[str] = None
        Name of the file where to save logs. Default: None. If None prints logs to the output.
    """
    if model_version is None:
        model_version = get_latest_model_version(model_name).version

    _get_logs(
        job_type=JobType.build,
        stream=stream,
        file_name=file_name,
        params={"model_name": model_name, "model_version": model_version},
    )


def get_venv_build_logs(
    model_name: str, model_version: Optional[int] = None, stream: bool = True, file_name: Optional[str] = None
) -> None:
    """
    Stream logs of the venv archive creating job by model name and version.

    Parameters
    ----------
    model_name: str
        The name of the model whose venv archive creation logs you want to view.
    model_version: Optional[int] = None
        The version of the model whose venv archive creation logs you want to view.
        Default: None, "latest" version is used.
    stream: bool = True
        Stream logs or dump all available at the moment.
    file_name: Optional[str] = None
        Name of the file where to save logs. Default: None. If None prints logs to the output.
    """
    if model_version is None:
        model_version = get_latest_model_version(model_name).version

    _get_logs(
        job_type=JobType.venv,
        stream=stream,
        file_name=file_name,
        params={"model_name": model_name, "model_version": model_version},
    )


def _get_logs(job_type: JobType, params: Dict[str, Any], stream: bool = True, file_name: Optional[str] = None) -> None:
    url = _get_logs_url(params=params, stream=stream, job_type=job_type)
    if stream:
        ws = None
        file = sys.stdout
        try:
            ws = AuthSession().instantiate_websocket_connection(url)
            if file_name:
                file = open(file_name, "w")
            while True:
                data = ws.recv()
                if not data:
                    return
                data = json.loads(data)
                if "status" not in data or data["status"] != "OK":
                    raise RuntimeError("Internal Server Error")
                print("\n".join(data["logs"]), file=file)
        except KeyboardInterrupt:
            pass
        except websocket._exceptions.WebSocketBadStatusException:
            possible_reasons_msg = (
                "Model with that name and version does not exist or does not have build job"
                if job_type is not JobType.execution
                else "Job with that name does not exist"
            )
            print(
                "<Connection refused. Check your query parameters, they may be incorrect.>\n"
                "<Possible reasons:>\n"
                f"<{possible_reasons_msg}>"
            )
        except Exception as err:
            print(err)
        finally:
            if ws is not None:
                ws.close()
            if file_name:
                file.close()
        return

    with AuthSession().get(url, stream=True) as resp:
        if file_name:
            with open(file_name, "a") as f:
                for line in resp.iter_lines():
                    if line:
                        f.writelines([line, "\n"])
        else:
            for line in resp.iter_lines():
                if line:
                    print(line)


def _get_logs_url(params: Dict[str, Any], job_type: JobType, stream: bool = True) -> str:
    server_url = get_server_websocket_url() if stream else get_server_url()
    url_base = posixpath.join(server_url, "logs-api")
    url_encoded_params = "&".join(f"{quote(str(key))}={quote(str(value))}" for key, value in params.items())
    local_path = f"{f'stream/{job_type.value}' if stream else f'filedump/{job_type.value}'}?{url_encoded_params}"
    return posixpath.join(url_base, local_path)


def add_ml_job(
    job_executor_name: str,
    executor_params: dict,
    models_pattern: Union[SingleModel, ArbitraryModels],
    data_pattern: Union[SingleDatasetLoader, ArbitraryDatasetLoaders],
    gpu: bool = False,
    job_executor_version: Optional[int] = None,
    experiment_name: str = "Default",
    cron_expression: Optional[str] = None,
    periodic_type: str = "ONCE",
    additional_system_packages: Optional[List[str]] = None,
) -> str:
    """
    Create execution job.

    Parameters
    ----------
    job_executor_name: str
        Name of the executor that will execute the job.
    executor_params: Dict[str, ...]
        Dictionary of executor parameters.
        Example::

            {
                'executor_param1': 'value1',
                'executor_param2': 'value2',
                'executor_param3': 'value3',
                ...
            }

    models_pattern: Union[SingleModel, ArbitraryModels]
        Necessary information for using the models.
    data_pattern: Union[SingleDatasetLoader, ArbitraryDatasetLoaders]
        Necessary information for using the datasets.
    gpu: bool = False
        Whether to use GPU for this job or not. Default: False
    job_executor_version: Optional[int] = None
        Version of the executor that will execute the job. Default: None, "latest" version is used.
    experiment_name: str = "Default"
        Name of the experiment. Default: "Default"
    cron_expression: str = None
        Cron expression for periodic or deferred jobs. Default: None
    periodic_type: {"ONCE", "PERIODIC"}
        Frequency of the task. Default: "ONCE"
    additional_system_packages: Optional[List[str]] = None
        List of system libraries for Debian family distributions that need to be installed in the job. Default: None

    Returns
    -------
    str
        Name of the Job.
    """
    if not isinstance(models_pattern, SingleModel) and not isinstance(models_pattern, ArbitraryModels):
        raise TypeError("Parameter models must have type SingleModel or ArbitraryModels.")
    if not isinstance(data_pattern, SingleDatasetLoader) and not isinstance(data_pattern, ArbitraryDatasetLoaders):
        raise TypeError("Parameter data_pattern must have type SingleDatasetLoader or ArbitraryDatasetLoaders.")
    models = models_pattern.serialize()
    data_params = data_pattern.serialize()
    list_role_data_params = []
    for data_param_role in data_params:
        op = Operation(schema.Query)
        data_inner_params: dict = data_param_role["data_params"]
        dataset_loader_name = data_inner_params["dataset_loader_name"]
        dataset_loader_version = data_inner_params.get("dataset_loader_version", None)
        dataset_loader_version_choice = schema.ObjectVersionOptionalInput(
            name=dataset_loader_name, version=dataset_loader_version
        )

        collector_name = data_inner_params["collector_name"]
        _dataset_loader_version = schema.ObjectVersionOptionalInput(
            name=dataset_loader_name, version=dataset_loader_version
        )
        dataset_collector_schema = op.dataset_loader_version_from_name_version(
            dataset_loader_version=_dataset_loader_version
        ).data_json_schema(collector_name=collector_name)
        dataset_collector_schema.collector_method_schema.schema_name()
        dataset_loader_version_obj = send_graphql_request(op, json_response=False)
        collector_method_name = (
            dataset_loader_version_obj.dataset_loader_version_from_name_version.data_json_schema.collector_method_schema.schema_name  # noqa: E501
        )
        collector_method_params = schema.MethodParamsInput(
            method_name=collector_method_name, method_params=json.dumps(data_inner_params["collector_params"])
        )

        list_dataset_loader_method_params = []
        for item in data_inner_params["dataset_loader_params"]:
            for key in item:
                list_dataset_loader_method_params.append(
                    schema.MethodParamsInput(method_name=key, method_params=json.dumps(item[key]))
                )

        current_dataset_loader_params = schema.DataParamsInput(
            dataset_loader_version_choice=dataset_loader_version_choice,
            collector_name=collector_name,
            list_dataset_loader_method_params=list_dataset_loader_method_params,
            collector_method_params=collector_method_params,
        )

        list_role_data_params.append(
            schema.RoleDataParamsInput(role=data_param_role["role"], data_params=current_dataset_loader_params)
        )

    op = Operation(schema.Query)
    _executor_version = schema.ObjectVersionOptionalInput(name=job_executor_name, version=job_executor_version)

    executor_version_choice = schema.ObjectVersionOptionalInput(name=job_executor_name, version=job_executor_version)

    executor_model_schema = op.executor_version_from_name_version(executor_version=_executor_version)
    executor_model_schema.executor_method_schema_name()

    executor_version_obj = send_graphql_request(op, json_response=False)
    executor_method_schema = executor_version_obj.executor_version_from_name_version.executor_method_schema_name

    executor_method_params = schema.MethodParamsInput(
        method_name=executor_method_schema, method_params=json.dumps(executor_params)
    )

    list_role_model_params = []

    for model in models:
        choice_criteria = model["model"].get("choice_criteria", "latest")
        version = model["model"].get("version")

        model_version_choice = schema.ModelVersionChoice(
            name=model["model"]["name"],
            version=version,
            choice_criteria=choice_criteria,
        )

        model_methods_params = []

        for item in model["params"]:
            for key in item:
                model_methods_params.append(
                    schema.MethodParamsInput(method_name=key, method_params=json.dumps(item[key]))
                )

        new_model_name = model.get("new_model_name")
        new_model_description = model.get("description")
        prepare_new_model_inference = model.get("prepare_new_model_inference", False)

        current_model_params = schema.ModelParamsInput(
            model_version_choice=model_version_choice,
            list_model_method_params=model_methods_params,
            new_model_name=new_model_name,
            prepare_new_model_inference=prepare_new_model_inference,
            description=new_model_description,
        )

        model_role = schema.RoleModelParamsInput(role=model["role"], model_params=current_model_params)

        list_role_model_params.append(model_role)

    executor_params = schema.ExecutorParamsInput(
        executor_method_params=executor_method_params,
        executor_version_choice=executor_version_choice,
    )

    op = Operation(schema.Mutation)
    mutation = op.add_ml_job(
        form=schema.JobParameters(
            executor_params=executor_params,
            list_role_model_params=list_role_model_params,
            list_role_data_params=list_role_data_params,
            experiment_name=experiment_name,
            cron_expression=cron_expression,
            periodic_type=periodic_type,
            gpu=gpu,
            additional_system_packages=additional_system_packages,
        )
    )

    mutation.name()

    job = send_graphql_request(op, json_response=False)

    return job.add_ml_job.name



import importlib
import json
import os
import posixpath
import subprocess
import sys
import tarfile
import threading
import warnings
from contextlib import _GeneratorContextManager
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Dict, List, Optional, Union

import cloudpickle
import httpx
import numpy
import pandas
import yaml
from scipy.sparse import csc_matrix, csr_matrix

import mlflow
from ML_management.loader.loader import CONFIG_KEY_ARTIFACTS
from ML_management.mlmanagement import variables
from ML_management.mlmanagement.base_exceptions import *  # noqa: F403
from ML_management.mlmanagement.base_exceptions import MLMClientError, MLMServerError, PylintError
from ML_management.mlmanagement.jsonschema_inference import infer_jsonschema
from ML_management.mlmanagement.mlmanager import set_experiment, start_run, start_run_if_not_exist
from ML_management.mlmanagement.model_type import ModelType
from ML_management.mlmanagement.module_finder import ModuleFinder
from ML_management.mlmanagement.server_mlmanager_exceptions import *  # noqa: F403
from ML_management.mlmanagement.server_mlmanager_exceptions import (
    AuthError,
    InvalidExperimentNameError,
    ModelTypeIsNotFoundError,
)
from ML_management.mlmanagement.session import AuthSession
from ML_management.mlmanagement.utils import INIT_FUNCTION_NAME, is_model_name_valid, validate_predict_config
from ML_management.mlmanagement.variables import (
    EXPERIMENT_NAME_FOR_DATASET_LOADER,
    EXPERIMENT_NAME_FOR_EXECUTOR,
    FILENAME_FOR_INFERENCE_CONFIG,
    active_run_stack,
    get_log_service_url,
    get_server_ml_api,
)
from ML_management.mlmanagement.visibility_options import VisibilityOptions
from ML_management.registry.exceptions import *  # noqa: F403
from mlflow.pyfunc import DATA


def _log_model(
    artifact_path,
    description: Optional[str],
    model_version_tags: Optional[Dict[str, str]] = None,
    code_path=None,
    conda_env=None,
    python_model=None,
    artifacts: Optional[dict] = None,
    registered_model_name: str = "default_name",
    signature: mlflow.models.signature.ModelSignature = None,
    input_example: Union[
        pandas.core.frame.DataFrame, numpy.ndarray, dict, list, csr_matrix, csc_matrix, str, bytes
    ] = None,
    await_registration_for: int = 300,
    pip_requirements=None,
    extra_pip_requirements=None,
    metadata=None,
    source_model_name=None,
    source_model_version=None,
    source_executor_name=None,
    source_executor_version=None,
    source_executor_role=None,
    upload_model_mode=None,
    visibility=None,
    extra_modules_names: Optional[list] = None,
    used_modules_names: Optional[list] = None,
    root_module_name: str = "__main__",
    linter_check: bool = True,
    start_build: bool = True,
    create_venv_pack: bool = False,
):
    """
    Log a Pyfunc model with custom inference logic and optional data dependencies as an MLflow artifact.

    Current run is using.
    Parameter registered_model_name must be not empty string,
        consist of alphanumeric characters, '_'
        and must start and end with an alphanumeric character.
        Validation regexp: "(([A-Za-z0-9][A-Za-z0-9_]*)?[A-Za-z0-9])+"
    You cannot specify the parameters: loader_module, data_path and the parameters: python_model, artifacts together.
    """
    from ML_management.executor.upload_model_mode import UploadModelMode  # circular import

    if extra_modules_names and used_modules_names:
        raise RuntimeError("Parameters 'extra_modules_names' and 'used_modules_names' cannot be used at the same time.")
    if upload_model_mode == UploadModelMode.none:
        raise RuntimeError("You can't log a model using the 'upload_model_mode' parameter set to none.")
    if not is_model_name_valid(registered_model_name):
        raise RuntimeError(
            "Parameter 'registered_model_name' must be not empty string, "
            "consist of alphanumeric characters, '_' "
            "and must start and end with an alphanumeric character."
            "Validation regexp: '(([A-Za-z0-9][A-Za-z0-9_]*)?[A-Za-z0-9])+'"
        )
    del UploadModelMode  # need to delete this because it is not JSON serializable
    if create_venv_pack:
        if not isinstance(artifacts, dict):
            raise RuntimeError("You must provide config for model serving in artifacts, if create_venv_pack=True")
        validate_predict_config(path=artifacts.get(FILENAME_FOR_INFERENCE_CONFIG))
    start_run_if_not_exist()
    if used_modules_names is not None:
        submodules = set(used_modules_names)
        ModuleFinder.import_modules_by_name(submodules)
    else:
        submodules = ModuleFinder().find_root_submodules(root_name=root_module_name)
        if linter_check:
            for module_name in submodules:
                if hasattr(sys.modules[module_name], "__file__"):
                    # pylint check what would fall on any error except when abstract method is not implemented (E0110)
                    # and when pylint has been unable to import a module (E0401)
                    linter_output = subprocess.run(
                        [
                            sys.executable,
                            "-m",
                            "pylint",
                            "{}".format(sys.modules[module_name].__file__),
                            "--clear-cache-post-run=True",
                            "--disable=all",
                            "--enable=E",
                            "--disable=E0401, E1101",
                            "--fail-on=E",
                            "--reports=n",
                            "--score=n",
                            "-j 0",
                        ],
                        check=False,
                        stdout=subprocess.PIPE,
                        text=True,
                    )
                    if linter_output.stdout:  # if linter has output when something wrong with code
                        raise PylintError(linter_output.stdout)
                    del linter_output
        if extra_modules_names:
            extra_set = set(extra_modules_names)
            ModuleFinder.import_modules_by_name(extra_set)
            submodules = submodules.union(extra_set)
    try:
        for module_name in submodules:
            cloudpickle.register_pickle_by_value(sys.modules[module_name])
        kwargs = {
            "artifact_path": artifact_path,
            "description": description,
            "model_version_tags": model_version_tags,
            "code_path": code_path,
            "conda_env": conda_env,
            "python_model": python_model,
            "registered_model_name": registered_model_name,
            "signature": signature,
            "input_example": input_example,
            "await_registration_for": await_registration_for,
            "pip_requirements": pip_requirements,
            "extra_pip_requirements": extra_pip_requirements,
            "metadata": metadata,
            "source_model_name": source_model_name,
            "source_model_version": source_model_version,
            "source_executor_name": source_executor_name,
            "source_executor_version": source_executor_version,
            "source_executor_role": source_executor_role,
            "upload_model_mode": upload_model_mode,
            "visibility": visibility,
            "start_build": start_build,
            "create_venv_pack": create_venv_pack,
        }
        response = _request_log_model("log_model", kwargs)
        return _raise_error(response)
    finally:
        for module_name in submodules:
            cloudpickle.unregister_pickle_by_value(sys.modules[module_name])


def _log_object_src(
    artifact_path,
    model_path: str,
    description: str,
    model_type: ModelType = ModelType.MODEL,
    model_version_tags: Optional[Dict[str, str]] = None,
    code_path=None,
    conda_env=None,
    python_model=None,
    registered_model_name: str = "default_name",
    signature: mlflow.models.signature.ModelSignature = None,
    input_example: Union[
        pandas.core.frame.DataFrame, numpy.ndarray, dict, list, csr_matrix, csc_matrix, str, bytes
    ] = None,
    await_registration_for: int = 300,
    pip_requirements=None,
    extra_pip_requirements=None,
    metadata=None,
    source_model_name=None,
    source_model_version=None,
    source_executor_name=None,
    source_executor_version=None,
    source_executor_role=None,
    upload_model_mode=None,
    visibility: VisibilityOptions = VisibilityOptions.PRIVATE,
    start_build: bool = True,
    create_venv_pack: bool = False,
    additional_local_packages: Optional[Union[List[str], str]] = None,
):
    """
    Log a Pyfunc model with custom inference logic and optional data dependencies as an MLflow artifact.

    Current run is using.
    Parameter registered_model_name must be not empty string,
            consist of alphanumeric characters, '_'
            and must start and end with an alphanumeric character.
            Validation regexp: "(([A-Za-z0-9][A-Za-z0-9_]*)?[A-Za-z0-9])+"
    You cannot specify the parameters: loader_module, data_path and the parameters: python_model, artifacts together.
    """
    from ML_management.executor.upload_model_mode import UploadModelMode  # circular import

    if not os.path.exists(model_path):
        raise FileNotFoundError(f"Path: {model_path} does not exist.")

    if not os.path.exists(os.path.join(model_path, "conda.yaml")):
        raise FileNotFoundError("There is no conda.yaml file.")

    try:
        with open(os.path.join(model_path, "conda.yaml")) as conda_file:
            dependencies = yaml.safe_load(conda_file)["dependencies"]  # noqa: F841
    except Exception:
        raise RuntimeError("conda.yaml not valid.") from None

    if upload_model_mode == UploadModelMode.none:
        raise RuntimeError("You can't log a model using the 'upload_model_mode' parameter set to none.")
    if not is_model_name_valid(registered_model_name):
        raise RuntimeError(
            "Parameter 'registered_model_name' must be not empty string, "
            "consist of alphanumeric characters, '_' "
            "and must start and end with an alphanumeric character."
            "Validation regexp: '(([A-Za-z0-9][A-Za-z0-9_]*)?[A-Za-z0-9])+'"
        )
    del UploadModelMode  # need to delete this because it is not JSON serializable
    if create_venv_pack:
        validate_predict_config(path=os.path.join(f"{model_path}", "artifacts", f"{FILENAME_FOR_INFERENCE_CONFIG}"))
    start_run_if_not_exist()
    python_path = sys.path
    try:
        model_path = os.path.abspath(model_path)
        parts = Path(model_path).parts
        sys.path.append(str(Path(*parts[:-2])))
        get_object_func = getattr(importlib.import_module(".".join(parts[-2:])), INIT_FUNCTION_NAME)
        python_model = get_object_func()  # noqa: F841
        kwargs = {
            "artifact_path": artifact_path,
            "model_path": model_path,
            "description": description,
            "model_version_tags": model_version_tags,
            "code_path": code_path,
            "conda_env": conda_env,
            "python_model": python_model,
            "registered_model_name": registered_model_name,
            "signature": signature,
            "input_example": input_example,
            "await_registration_for": await_registration_for,
            "pip_requirements": pip_requirements,
            "extra_pip_requirements": extra_pip_requirements,
            "metadata": metadata,
            "source_model_name": source_model_name,
            "source_model_version": source_model_version,
            "source_executor_name": source_executor_name,
            "source_executor_version": source_executor_version,
            "source_executor_role": source_executor_role,
            "upload_model_mode": upload_model_mode,
            "visibility": visibility,
            "start_build": start_build,
            "create_venv_pack": create_venv_pack,
            "additional_local_packages": additional_local_packages,
            "get_object_func": get_object_func,
        }
        response = _request_log_model("log_object_src", kwargs)
        result = _raise_error(response)
    except Exception as err:
        raise err
    finally:
        sys.path = python_path

    if additional_local_packages:
        if isinstance(additional_local_packages, list):
            for package in additional_local_packages:
                response = _request_log_artifacts(package, DATA, model_type)
                _raise_error(response)
        else:
            response = _request_log_artifacts(additional_local_packages, DATA, model_type)
            _raise_error(response)
    return result


def log_executor_src(
    model_path: str,
    registered_name: str,
    description: str,
    start_build: bool = False,
    visibility: VisibilityOptions = VisibilityOptions.PRIVATE,
    additional_local_packages: Optional[Union[List[str], str]] = None,
):
    """
    Upload executor using folder with source code of the executor.

    model_path folder must contain "__init__.py" file with get_object function
    which have to return instance of the executor.
    Also model_path must contain "conda.yaml" file with dependencies of the executor
    model_path can optionally contain "artifacts" folder with files/folders,
    which would be automatically uploaded with executor
    and path to them can be accessed with self.artifacts within executor class.

    Parameters
    ----------
    model_path: str
        Path to folder with executor.
    registered_name: str
        Name of the executor.
    description: str
        Description of the executor
    start_build: bool
        Is to start build image of the executor right after it was logged. Defaults to False.
    visibility: VisibilityOptions
        Visibility of this executor to other users. Possible values VisibilityOptions.PRIVATE, PUBLIC.
        Defaults to PRIVATE.
    additional_local_packages: Optional[Union[List[str], str]] = None
        Path or list of paths to folder with local dependencies which are not within model_path folder.
        Defaults to None.

    Returns
    -------
    None

    """
    old_experiment_name = variables.active_experiment

    try:
        set_experiment(EXPERIMENT_NAME_FOR_EXECUTOR, visibility=VisibilityOptions.PUBLIC)

        with start_run(nested=True):
            _log_object_src(
                artifact_path="",
                description=description,
                registered_model_name=registered_name,
                start_build=start_build,
                model_path=model_path,
                visibility=visibility,
                additional_local_packages=additional_local_packages,
                model_type=ModelType.EXECUTOR,
            )

    except Exception as err:
        raise err
    finally:
        variables.active_experiment = old_experiment_name


def log_dataset_loader_src(
    model_path: str,
    registered_name: str,
    description: str,
    visibility: VisibilityOptions = VisibilityOptions.PRIVATE,
    additional_local_packages: Optional[Union[List[str], str]] = None,
):
    """
    Upload dataset loader using folder with source code of the dataset loader.

    model_path folder must contain "__init__.py" file with get_object function
    which have to return instance of the dataset loader.
    Also model_path must contain "conda.yaml" file with dependencies of the dataset loader
    model_path can optionally contain "artifacts" folder with files/folders,
    which would be automatically uploaded with dataset loader
    and path to them can be accessed with self.artifacts within dataset loader class.

    Parameters
    ----------
    model_path: str
        Path to folder with dataset loader.
    registered_name: str
        Name of the dataset loader.
    description: str
        Description of the dataset loader
    visibility: VisibilityOptions
        Visibility of this dataset loader to other users. Possible values VisibilityOptions.PRIVATE, PUBLIC.
        Defaults to PRIVATE.
    additional_local_packages: Optional[Union[List[str], str]] = None
        Path or list of paths to folder with local dependencies which are not within model_path folder.
        Defaults to None.

    Returns
    -------
    None

    """
    old_experiment_name = variables.active_experiment
    try:
        set_experiment(EXPERIMENT_NAME_FOR_DATASET_LOADER, visibility=VisibilityOptions.PUBLIC)
        with start_run(nested=True):
            _log_object_src(
                artifact_path="",
                description=description,
                registered_model_name=registered_name,
                model_path=model_path,
                visibility=visibility,
                additional_local_packages=additional_local_packages,
                model_type=ModelType.DATASET_LOADER,
            )
    except Exception as err:
        raise err
    finally:
        variables.active_experiment = old_experiment_name


def log_model_src(
    model_path: str,
    registered_name: str,
    description: str,
    start_build: bool = True,
    model_version_tags: Optional[Dict[str, str]] = None,
    experiment_name: Optional[str] = None,
    create_venv_pack: bool = False,
    visibility: VisibilityOptions = VisibilityOptions.PRIVATE,
    additional_local_packages: Optional[Union[List[str], str]] = None,
) -> None:
    """
    Log model to remote server.

    Parameters
    ----------
    model_path: str
        Path to the folder with the model files(*model*.py, conda.yaml, __init__.py with get object function, see docs)
    registered_name: str
        The name of the model that will be assigned to the model in the model registry.
        Parameter registered_name should match regexp ``"(([A-Za-z0-9][A-Za-z0-9_]*)?[A-Za-z0-9])+"``.
    description: str
        Desciption of the model.
    visibility: VisibilityOptions
        Visibility of this model to other users. Possible values VisibilityOptions.PRIVATE, PUBLIC.
        Defaults to PRIVATE.
    start_build: bool = True
        Whether to start building the image with the model requirements immediately after uploading.
        This can speed up the subsequent build of the task image. Default: True.
    model_version_tags: Optional[Dict[str, str]] = None
        Define model version tags. Default: None.
    create_venv_pack: bool = False
        Whether to prepare venv pack for future inference. Default: False.
    additional_local_packages: Optional[Union[List[str], str]] = None
        Path or list of paths to folder with local dependencies which are not within model_path folder.
        Defaults to None.

    Returns
    =======
    None
    """
    old_experiment_name = variables.active_experiment
    try:
        if experiment_name:
            set_experiment(experiment_name, visibility=VisibilityOptions.PUBLIC)
        with start_run(nested=True):
            _log_object_src(
                artifact_path="",
                description=description,
                registered_model_name=registered_name,
                model_path=model_path,
                start_build=start_build,
                model_version_tags=model_version_tags,
                create_venv_pack=create_venv_pack,
                visibility=visibility,
                additional_local_packages=additional_local_packages,
                model_type=ModelType.MODEL,
            )
    except Exception as err:
        raise err
    finally:
        variables.active_experiment = old_experiment_name


def log_artifact(local_path: str, artifact_path: Optional[str] = None) -> None:
    """
    Log a local file or directory as an artifact of the currently active run.

    If no run is active, this method will create a new active run.

    Parameters
    ==========
    local_path: str
        Path to the file to write.
    artifact_path: Optional[str] = None
        If provided, the directory to write to.

    Returns
    =======
    None
    """
    response = _request_log_artifacts(local_path, artifact_path)
    return _raise_error(response)


def _log_param(key: str, value: str) -> None:
    """
    Log str parameter (e.g. model hyperparameter) under the current job.

    Parameters
    ==========
    key: str
        Param name (string).
        This string may only contain alphanumerics, underscores (_), dashes (-),
        periods (.), spaces ( ), and slashes (/).
        All backend stores will support keys up to length 250, but some may support larger keys.

    value: str
        Param value (string).

    Returns
    =======
    None
    """
    if not isinstance(value, str):
        warnings.warn(f"The log_param function can log only string values. Value {value} is not log.")
        return
    return _request_log_params({key: value})


def _log_params(params: Dict[str, str]) -> None:
    """
    Log a batch of params for the current job.

    Parameters
    ==========
    params: dict
        Key parameter value pairs.
        Кey is the name of the parameter (string).
        This string may only contain alphanumerics, underscores (_), dashes (-),
        periods (.), spaces ( ), and slashes (/).
        All backend stores will support keys up to length 250, but some may support larger keys.
        Param value (string).

    Returns
    =======
    None
    """
    for key, value in params:
        if not isinstance(value, str):
            warnings.warn(f"The log_params function can log only string values. Value {value} is not log.")
        params.pop(key)
    if not params:
        return
    return _request_log_params(params)


def log_metric(key: str, value: float, step: int = 0) -> None:
    """
    Log a metric under the current job.

    Parameters
    ==========
    key: str
        Metric name (string).
        This string may only contain alphanumerics, underscores (_), dashes (-),
        periods (.), spaces ( ), and slashes (/).
        All backend stores will support keys up to length 250, but some may support larger keys.

    value: float
        Metric value (float). Note that some special values such as +/-
        Infinity may be replaced by other values depending on the store.
        For example, the SQLAlchemy store replaces +/- Infinity with max / min float values.
        All backend stores will support values up to length 5000, but some may support larger values.

    step: int
        Metric step (int). Defaults to zero if unspecified.

    Returns
    =======
    None
    """
    if not (isinstance(value, float) or isinstance(value, int)):
        warnings.warn(f"The log_metric function can log only float or integer values. Value {value} is not log.")
        return
    return _request_log_metric({key: value}, step)


def log_metrics(metrics: Dict[str, float], step: int = 0) -> None:
    """
    Log a metric under the current job.

    Parameters
    ==========
    metrics: dict[str, float]
        Key(string): metric name.
        This string may only contain alphanumerics, underscores (_), dashes (-),
        periods (.), spaces ( ), and slashes (/).
        All backend stores will support keys up to length 250, but some may support larger keys.
        Value(float): metric value. Note that some special values such as +/-
        Infinity may be replaced by other values depending on the store.
        For example, the SQLAlchemy store replaces +/- Infinity with max / min float values.
        All backend stores will support values up to length 5000, but some may support larger values.

    step: int
        Metric step (int). Defaults to zero if unspecified.

    Returns
    =======
    None
    """
    metrics_to_log = {}
    for key, value in metrics.items():
        if not (isinstance(value, float) or isinstance(value, int)):
            warnings.warn(f"The log_metrics function can log only float or integer values. {key}: {value} is not log.")
            continue
        if not isinstance(key, str):
            warnings.warn(
                f"The log_metric function can log only str name metric and float or integer values. "
                f"{key}: {value} is not log."
            )
            continue
        metrics_to_log[key] = value
    return _request_log_metric(metrics_to_log, step)


def _request_log_model(function_name: str, kwargs: dict):
    """
    Send request for log_model function.

    Steps for log model:
    0) Infer jsonschema, raise if it is invalid
    1) open temporary directory
    2) Do mlflow.save_model() locally
    3) Pack it to tar file
    4) Send it to server to log model there.
    """
    delete_args_for_save_model_func = [
        "description",
        "model_version_tags",
        "artifact_path",
        "registered_model_name",
        "await_registration_for",
        # now, extra arguments
        "upload_model_mode",
        "source_model_name",
        "source_model_version",
        "visibility",
        "source_executor_name",
        "source_executor_version",
        "source_executor_role",
        "start_build",
        "create_venv_pack",
    ]  # not need for save_model

    extra_imports_args = [
        "submodules",
        "module_name",
        "used_modules_names",
        "extra_modules_names",
        "root_module_name",
        "linter_check",
    ]

    delete_args_for_log_func = [
        "python_model",
        "artifacts",
        "conda_env",
        "pip_requirements",
        "extra_pip_requirements",
        "additional_local_packages",
        "conda_file",
        "dependencies",
        "get_object_func",
    ]  # not need for log model on server

    for delete_arg in extra_imports_args:
        kwargs.pop(delete_arg, None)
    kwargs_for_save_model = kwargs.copy()
    for delete_arg in delete_args_for_save_model_func:
        kwargs_for_save_model.pop(delete_arg, None)
    python_model = kwargs_for_save_model["python_model"]
    get_object_func = kwargs_for_save_model.get("get_object_func")
    # import some modules here because of circular import
    from ML_management.dataset_loader.dataset_loader_pattern import DatasetLoaderPattern
    from ML_management.dataset_loader.dataset_loader_pattern_to_methods_map import (
        dataset_loader_pattern_to_methods,
    )
    from ML_management.executor.base_executor import BaseExecutor
    from ML_management.executor.executor_pattern_to_methods_map import executor_pattern_to_methods
    from ML_management.model.model_type_to_methods_map import model_pattern_to_methods
    from ML_management.model.patterns.model_pattern import Model

    experiment_name = variables.active_experiment.get("experiment_name")

    if python_model is not None:
        if isinstance(python_model, Model):
            kwargs["model_type"] = ModelType.MODEL
            model_to_methods = model_pattern_to_methods
            if experiment_name in [
                EXPERIMENT_NAME_FOR_EXECUTOR,
                EXPERIMENT_NAME_FOR_DATASET_LOADER,
            ]:
                raise InvalidExperimentNameError(ModelType.MODEL.value, experiment_name)
        elif isinstance(python_model, BaseExecutor):
            kwargs["model_type"] = ModelType.EXECUTOR
            model_to_methods = executor_pattern_to_methods

            if experiment_name != EXPERIMENT_NAME_FOR_EXECUTOR:
                raise InvalidExperimentNameError(ModelType.EXECUTOR.value, experiment_name)
            # collect all needed model's methods
            kwargs["desired_model_methods"] = python_model.desired_model_methods
            kwargs["upload_model_modes"] = python_model.upload_model_modes
            kwargs["desired_dataset_loader_methods"] = python_model.desired_dataset_loader_methods
        elif isinstance(python_model, DatasetLoaderPattern):
            kwargs["model_type"] = ModelType.DATASET_LOADER
            model_to_methods = dataset_loader_pattern_to_methods
            if experiment_name != EXPERIMENT_NAME_FOR_DATASET_LOADER:
                raise InvalidExperimentNameError(kwargs["model_type"].value, experiment_name)
        else:
            raise ModelTypeIsNotFoundError()

        # now we need to infer schemas for methods.
        methods_schema = {}
        for model_type, methods_name_to_schema_map in model_to_methods.items():
            if isinstance(python_model, model_type):
                for method_name_to_schema in methods_name_to_schema_map:
                    model_method = getattr(python_model, method_name_to_schema.value, None)
                    model_method_schema = infer_jsonschema(model_method, get_object_func)
                    methods_schema[method_name_to_schema.value] = model_method_schema

        kwargs["model_method_schemas"] = methods_schema

        for delete_arg in delete_args_for_log_func:
            kwargs.pop(delete_arg, None)

        log_request = {
            "function_name": function_name,
            "active_experiment": variables.active_experiment,
            "active_run_ids": [run.info.run_id for run in active_run_stack],
        }

        if function_name == "log_model":
            kwargs["loader_module"] = mlflow.pyfunc.model.__name__
            with TemporaryDirectory() as temp_dir:
                model_folder = "model"
                path_for_model_folder = os.path.join(temp_dir, model_folder)
                mlflow.pyfunc.save_model(path=path_for_model_folder, **kwargs_for_save_model)
                model_folder = path_for_model_folder
                log_request["kwargs"] = kwargs
                return _open_pipe_send_request(model_folder, log_request, url=get_log_service_url("log_model"))
        else:
            artifacts_path = os.path.join(kwargs["model_path"], CONFIG_KEY_ARTIFACTS)
            if os.path.isfile(artifacts_path):
                raise Exception(f"The artifact file {artifacts_path} is invalid. The artifact must be a directory.")

            model_folder = kwargs["model_path"]

            del kwargs["model_path"]
            kwargs["loader_module"] = "ML_management.loader.loader"
            log_request["kwargs"] = kwargs
            return _open_pipe_send_request(model_folder, log_request, url=get_log_service_url("log_model"))

    else:
        raise Exception("python_model parameter must be specified")


def _request_log_artifacts(local_path, artifact_path, model_type: Optional[ModelType] = None):
    """Send request for log artifact."""
    log_artifact_request = {
        "kwargs": {"local_path": local_path, "artifact_path": artifact_path},
        "active_run_ids": [run.info.run_id for run in active_run_stack],
        "model_type": model_type,
    }
    if not os.path.exists(local_path):
        raise FileNotFoundError(f"Path: {local_path} does not exist.")
    if not os.path.isdir(local_path):
        basename = os.path.basename(os.path.normpath(local_path))
        with open(local_path, "rb") as file:
            url = get_log_service_url("log_artifact")

            # upload multipart
            data = {"mlflow_request": json.dumps(log_artifact_request)}
            headers = {"Transfer-Encoding": "chunked"}

            file_content_type = "application/octet-stream"

            files = {"file": (basename, file, file_content_type)}

            with AuthSession().post(
                url=url,
                stream=False,
                data=data,
                files=files,
                headers=headers,
            ) as response:
                return response

    return _open_pipe_send_request(local_path, log_artifact_request, url=get_log_service_url("log_artifact"))


def _raise_error(response: httpx.Response):
    if response.status_code == 500:
        raise MLMServerError("Internal server error.")
    if response.status_code != 200:
        detail = response.read().decode()
        if not detail:
            raise MLMServerError("Internal server error.")
        try:
            detail = json.loads(detail).get("detail")
        except Exception:
            raise MLMServerError(f"Server error '{detail}' with code {response.status_code}") from None
        if not (
            isinstance(detail, dict) and "exception_class" in detail and ("params" in detail or "message" in detail)
        ):
            raise MLMServerError(detail)
        if "params" in detail:
            error = getattr(sys.modules[__name__], detail["exception_class"])(**detail["params"])
        else:
            error = getattr(sys.modules[__name__], detail["exception_class"])(detail["message"])
        if isinstance(error, AuthError):
            error.args = (
                f"{error.args[0]}. "
                "Possible reason: you are trying to upload a version of an object owned by another user.",
            )
        raise error


def _open_pipe_send_request(folder, request_dict, url):
    r, w = os.pipe()

    try:
        thread = threading.Thread(target=_tar_folder, args=(w, folder))
        thread.start()
    except Exception as err:
        os.close(r)
        os.close(w)
        raise err

    with open(r, "rb") as buff:
        with _request(url, request_dict, buff, os.path.basename(folder), True) as response:
            return response


def _request(
    url,
    request_dict,
    file=None,
    basename=None,
    is_tar=False,
    stream=False,
) -> _GeneratorContextManager:
    """Create mlflow_request and send it to server."""
    if not file:
        return AuthSession().post(url=url, stream=stream, json=request_dict)

    # upload multipart
    data = {"mlflow_request": json.dumps(request_dict)}
    headers = {"Transfer-Encoding": "chunked"}

    file_content_type = "application/octet-stream"
    if is_tar:
        file_content_type = "application/x-tar"

    files = {"file": (basename, file, file_content_type)}

    return AuthSession().post(
        url=url,
        stream=stream,
        data=data,
        files=files,
        headers=headers,
    )


def _tar_folder(w, model_folder):
    try:
        with open(w, "wb") as buff:
            with tarfile.open(mode="w|", fileobj=buff) as tar:
                tar.add(model_folder, arcname=os.path.basename(model_folder))
    except Exception as err:
        raise MLMClientError("Some error during tar the content.") from err


def _request_log_metric(metrics: Dict[str, float], step: Optional[int] = None):
    if not active_run_stack:
        raise MLMClientError("The log_metric function must be called from the active job.")
    run_id = active_run_stack[0].info.run_id

    request_dict = {"metrics": metrics, "step": step, "run_id": run_id}

    url = posixpath.join(get_server_ml_api(), "log-metric")

    _request_with_warning(url, request_dict)


def _request_log_params(params: Dict[str, str]):
    if not active_run_stack:
        raise MLMClientError("The log_params function must be called from the active job.")
    run_id = active_run_stack[0].info.run_id

    request_dict = {"params": params, "run_id": run_id}

    url = posixpath.join(get_server_ml_api(), "log-params")

    _request_with_warning(url, request_dict)


def _request_with_warning(url, request_dict):
    with AuthSession().post(url=url, json=request_dict) as response:
        response_content = response.content
        if response.status_code != 200:
            warnings.warn(f"Server error: {response_content.decode()}, status: {response.status_code}")
            return
